"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np

from qrisp.operators import Hamiltonian
from qrisp.operators.fermionic.fermionic_term import FermionicTerm
from qrisp.operators.hamiltonian_tools import group_up_iterable
from qrisp import merge, IterationEnvironment, conjugate
from qrisp.operators.qubit import QubitOperator

from qrisp.jasp import check_for_tracing_mode, jrange
import sympy as sp

threshold = 1e-9


def fermionic_trotterization(H, forward_evolution=True):

    H = H.hermitize()
    reduced_H = H.reduce(assume_hermitian=True)
    groups = reduced_H.group_up(denominator=lambda a, b: a.unipolars_agree(b))

    def meta_group_denominator(H0, H1):
        term_0 = next(iter(H0.terms_dict.keys()))
        term_1 = next(iter(H1.terms_dict.keys()))
        return not term_0.unipolars_intersect(term_1)

    meta_groups = group_up_iterable(groups, meta_group_denominator)
    groups = [sum(meta_group, 0) for meta_group in meta_groups]

    def trotter_step(qarg, t, steps):

        n = H.find_minimal_qubit_amount()
        qarg = [qarg[i] for i in range(n)]

        for group in groups:

            # We now treat the fermionic swaps.
            # The problem here is that terms like
            # a(0)*a(2) + a(1)*a(3)
            # Have the JW embedding
            # -A(0)*Z(1)*A(2) - A(1)*Z(2)*A(3)
            # Implying the both need access to qubit 1&2 which makes them block
            # each other.
            # The goal is therefore to reorder the terms via fermionic swaps
            # to unblock. For an overview over the fermionic swapping topic
            # please check https://arxiv.org/abs/2310.12256

            # Obviously we want to reduce the amount of fermionic swaps to a minimum.
            # We approach this by noticing that (contrary to Selingers approach),
            # it is not necessary to group all ladder operators together. It is
            # sufficient to match them into "couples".
            # a(0)*a(1)*a(7)*a(8)
            # => A(0)*A(1)*A(7)*A(8)

            # Matching the ladder operator into couples takes significantly less
            # swaps them grouping them up in one big chunk.

            # For that reason we will now go through the ladder terms that need
            # to be simulated and match them into couples and singles.
            # Singles are ladder terms that can't be matched because the corresponding
            # operator targets an odd amount of qubits.
            # Consider for instance
            # a(1)*a(3)*a(4)
            # Here, 3&4 are a couple and 1 is a single

            terms = list(group.terms_dict.keys())

            singles = []
            couples = {}
            n = group.find_minimal_qubit_amount()

            for term in terms:

                # For non-unipolar factors (i.e. a(i)*c(i) for instance), no matching is necessary.
                index_list = term.get_unipolars()

                # If there is an od amount of ladder operators, remove the last one
                # (has the lowest index)
                if len(index_list) % 2:
                    singles.append(index_list.pop(-1))

                # We now group the ladder operators into couples
                for i in range(len(index_list) // 2):
                    couples[index_list[2 * i + 1]] = index_list[2 * i]

            # This function computes the swaps that are necessary to match
            # all couples and moves the singles to the lowest positions.
            swaps, permutation = kai_pflaume(singles, couples, n)

            # This function applies the CZ gates on the quantum argument to
            # perform the fermionic swap
            with conjugate(apply_fermionic_swaps)(qarg, swaps) as new_qarg:
                for ferm_term in terms:
                    coeff = reduced_H.terms_dict[ferm_term]

                    # This function permutes the indices of the fermionic term
                    qubit_operator = ferm_term.fermionic_swap(
                        permutation
                    ).to_qubit_term()
                    qubit_term = list(qubit_operator.terms_dict.keys())[0]

                    if not len(ferm_term.get_unipolars()) % 2:
                        for factor in qubit_term.factor_dict.values():
                            if factor == "Z":
                                raise Exception(
                                    "Fermionic matching failed: Z Operator found"
                                )

                    qubit_term.simulate(
                        -coeff
                        * t
                        / steps
                        * qubit_operator.terms_dict[qubit_term]
                        * (-1) ** int(forward_evolution),
                        new_qarg,
                    )

    def U(qarg, t=1, steps=1, iter=1):
        if check_for_tracing_mode():
            for i in jrange(iter * steps):
                trotter_step(qarg, t, steps)
        else:
            merge([qarg])
            with IterationEnvironment(qarg.qs, iter * steps):
                trotter_step(qarg, t, steps)

    return U


# This function takes a list of indices (singles) and a dictionary of indices (couples)
# and generates the swaps to move the K singles to the K lowest positions. It furthermore
# generates the swaps to move all couples adjacent to each other.
def kai_pflaume(singles, couples, n):

    permutation = list(range(n))
    swaps = []

    singles.sort()

    k = 0
    for s in singles:
        # Move the k-th single to position k
        for i in range(k, s)[::-1]:
            swap = (i + 1, i)
            permutation[swap[0]], permutation[swap[1]] = (
                permutation[swap[1]],
                permutation[swap[0]],
            )
            swaps.append(swap)
        k += 1

    # The female indices are the indices that are moved towards the males
    # Imagine we have the female 3 and the male 6
    # We start with the permutation
    # [0,1,2,3,4,5,6,7]
    # We now need to move 3 adjacent to 6 and record the necessary swaps
    # We end up in
    # [0,1,2,4,5,3,6,7]

    # We begin by moving the female with the highest index first to avoid
    # accidentally moving other females away from their partner.
    females = list(couples.keys())
    females.sort()
    females.reverse()

    for f in females:

        m = couples[f]
        # Move the female adjacent to the male
        for i in range(permutation.index(f), permutation.index(m) - 1):
            swap = (i, i + 1)
            permutation[swap[0]], permutation[swap[1]] = (
                permutation[swap[1]],
                permutation[swap[0]],
            )
            swaps.append(swap)

    return swaps, permutation


def apply_fermionic_swaps(qarg, swaps):
    from qrisp import cz

    qb_list = list(qarg)

    for swap in swaps:
        cz(qb_list[swap[0]], qb_list[swap[1]])
        qb_list[swap[0]], qb_list[swap[1]] = qb_list[swap[1]], qb_list[swap[0]]

    return qb_list


def apply_fermionic_swap(qv, permutation):
    from qrisp import cz

    qb_list = list(qv)
    swaps = get_swaps_for_permutation(permutation)
    for swap in swaps[::-1]:
        cz(qb_list[swap[0]], qb_list[swap[1]])
        qb_list[swap[0]], qb_list[swap[1]] = qb_list[swap[1]], qb_list[swap[0]]

    return qb_list


def get_swaps_for_permutation(permutation):
    swaps = []
    permutation = list(permutation)
    for i in range(len(permutation)):
        j = permutation.index(i)
        while j != i:
            permutation[j], permutation[j - 1] = permutation[j - 1], permutation[j]
            swaps.append((j, j - 1))
            j -= 1
    return swaps



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import weakref

import jax

from qrisp.jasp.primitives import create_qubits, delete_qubits_p, OperationPrimitive
from qrisp.jasp.tracing_logic.dynamic_qubit_array import DynamicQubitArray
from qrisp.core.quantum_variable import QuantumVariable

from sympy import symbols

greek_letters = symbols(
    "alpha beta gamma delta epsilon zeta eta theta iota kappa lambda mu nu xi omicron pi rho sigma tau upsilon phi chi psi omega"
)


class TracingQuantumSession:
    tr_qs_container = [None]
    abs_qc_stack = []
    qubit_cache_stack = []

    def __init__(self):

        self.abs_qc = None
        self.qv_list = []
        self.deleted_qv_list = []
        self.qubit_cache = None
        TracingQuantumSession.tr_qs_container.insert(0, self)
        self.qv_stack = []
        self.gc_mode_stack = []
        self.gc_mode = "auto"

    def start_tracing(self, abs_qc, gc_mode=None):
        self.abs_qc_stack.append(self.abs_qc)
        self.qubit_cache_stack.append(self.qubit_cache)

        self.abs_qc = abs_qc
        self.qubit_cache = {}

        self.qv_stack.append((self.qv_list, self.deleted_qv_list))
        self.qv_list = []
        self.deleted_qv_list = []

        self.gc_mode_stack.append(self.gc_mode)
        if gc_mode is None:
            self.gc_mode = self.gc_mode_stack[-1]
        else:
            self.gc_mode = gc_mode

    def garbage_collection(self, spare_qv_list):

        if self.gc_mode in ["auto", "debug"]:
            from qrisp import reset

            spare_hash_list = [hash(qv) for qv in spare_qv_list]
            for qv in list(self.qv_list):
                if hash(qv) in spare_hash_list:
                    continue
                if self.gc_mode == "auto":
                    reset(qv)
                    qv.delete()
                else:
                    raise Exception(
                        f"QuantumVariable {qv} went out of scope without deallocation"
                    )

    def conclude_tracing(self):

        temp = self.abs_qc
        self.abs_qc = self.abs_qc_stack.pop(-1)
        self.qubit_cache = self.qubit_cache_stack.pop(-1)
        self.qv_list, self.deleted_qv_list = self.qv_stack.pop(-1)
        self.gc_mode = self.gc_mode_stack.pop(-1)

        return temp

    def append(self, operation, qubits=[], clbits=[], param_tracers=[]):
        
        if not self.abs_qc._trace is jax.core.trace_ctx.trace:
            raise Exception(
                """Lost track of QuantumCircuit during tracing. This might have been caused by a missing quantum_kernel decorator or not using quantum prefix control (like q_fori_loop, q_cond). Please visit https://www.qrisp.eu/reference/Jasp/Quantum%20Kernel.html for more details"""
            )

        if len(clbits):
            raise Exception(
                "Tried to append Operation with non-zero classical bits in JAX mode."
            )

        from qrisp.core import QuantumVariable, QuantumArray
        from qrisp.jasp import jrange

        if isinstance(qubits[0], (QuantumVariable, DynamicQubitArray)):

            for i in jrange(qubits[0].size):
                self.append(
                    operation,
                    [qubits[j][i] for j in range(operation.num_qubits)],
                    param_tracers=param_tracers,
                )
            return

        elif isinstance(qubits[0], list):

            for i in range(len(qubits[0])):
                self.append(
                    operation,
                    [qubits[j][i] for j in range(operation.num_qubits)],
                    param_tracers=param_tracers,
                )
            return
        
        elif isinstance(qubits[0], QuantumArray):
            
            for i in range(1, len(qubits)):
                if not isinstance(qubits[i], QuantumArray):
                    raise Exception(f"Tried to apply multi-qubit gate to mixed qubit argument types (QuantumArray + {type(qubits[i])})")
                
                if qubits[i].shape != qubits[0].shape:
                    raise Exception("Tried to apply multi-qubit quantum gate to QuantumArrays of differing shape.")
            
            flattened_qubits = [qubits[i].flatten() for i in range(len(qubits))]
            
            for i in jrange(flattened_qubits[0].size):
                self.append(
                    operation,
                    [flattened_qubits[j][i] for j in range(operation.num_qubits)],
                    param_tracers=param_tracers,
                )
            return

        temp_op = operation.copy()

        temp_op.params = list(temp_op.params)
        for i in range(len(temp_op.params)):
            temp_op.params[i] = greek_letters[i]

        op_primitive = OperationPrimitive(temp_op)

        self.abs_qc = op_primitive.bind(
            *([b for b in qubits] + param_tracers + [self.abs_qc])
        )

    def register_qv(self, qv, size):

        if not self.abs_qc._trace is jax.core.trace_ctx.trace:
            raise Exception(
                """Lost track of QuantumCircuit during tracing. This might have been caused by a missing quantum_kernel decorator or not using quantum prefix control (like q_fori_loop, q_cond). Please visit https://www.qrisp.eu/reference/Jasp/Quantum%20Kernel.html for more details"""
            )

        # Determine amount of required qubits
        if size is not None:
            qv.reg = self.request_qubits(size)
            
        # Register in the list of active quantum variable
        self.qv_list.append(qv)
        qv.qs = self

        QuantumVariable.live_qvs.append(weakref.ref(qv))
        qv.creation_time = int(QuantumVariable.creation_counter[0])
        QuantumVariable.creation_counter += 1
        
    def request_qubits(self, amount):
        qb_array_tracer, self.abs_qc = create_qubits(amount, self.abs_qc)
        return DynamicQubitArray(qb_array_tracer)
        

    def delete_qv(self, qv, verify=False):
        
        if not self.abs_qc._trace is jax.core.trace_ctx.trace:
            raise Exception(
                """Lost track of QuantumCircuit during tracing. This might have been caused by a missing quantum_kernel decorator or not using quantum prefix control (like q_fori_loop, q_cond). Please visit https://www.qrisp.eu/reference/Jasp/Quantum%20Kernel.html for more details"""
            )

        if verify == True:
            raise Exception("Tried to verify deletion in tracing mode.")

        # Check if quantum variable appears in this session
        if qv.name not in [qv.name for qv in self.qv_list]:
            raise Exception(
                "Tried to remove a non existent quantum variable from quantum session"
            )

        self.clear_qubits(qv.reg, verify)

        # Remove quantum variable from list
        for i in range(len(self.qv_list)):
            temp_qv = self.qv_list[i]

            if temp_qv.name == qv.name:
                self.qv_list.pop(i)
                break

        self.deleted_qv_list.append(qv)

    def clear_qubits(self, qubits, verify=False):

        self.abs_qc = delete_qubits_p.bind(qubits.tracer, self.abs_qc)

    @classmethod
    def release(cls):
        cls.tr_qs_container.pop(0)

    @classmethod
    def get_instance(cls):
        return cls.tr_qs_container[0]


tracing_qs_singleton = TracingQuantumSession()


def check_for_tracing_mode():
    return hasattr(jax._src.core.trace_ctx.trace, "frame")

def get_last_equation(i = -1):
    return jax._src.core.trace_ctx.trace.frame.eqns[i]

def check_live(tracer):
    if tracer is None:
        return True
    return not not tracer._trace.main.jaxpr_stack



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.circuit import Qubit
from qrisp.qtypes import QuantumFloat
from qrisp.core.gate_application_functions import x, cx, mcx, swap

def peres(a : Qubit, b : Qubit, c : Qubit):
    mcx([a, b], c)
    cx(a, b)

def add(A: list[Qubit], B: list[Qubit]):
    n = len(A)
    # Step 1
    for i in range(1, n, 1):
        cx(B[i], A[i])
        
    # Step 2
    for i in range(n-2, 0, -1):
        cx(B[i], B[i+1])

    # Step 3
    for i in range(0, n-1, 1):
        mcx([B[i], A[i]], B[i+1])

    # Step 4
    for i in range(n-1, -1, -1):
        if(i == n - 1):
            cx(B[i], A[i])
        else:
            peres(B[i], A[i], B[i+1])

    # Step 5
    for i in range(1, n-1, 1):
        cx(B[i], B[i+1])

    # Step 6
    for i in range(1, n, 1):
        cx(B[i], A[i])

def control_add(z: Qubit, B: list[Qubit], A: list[Qubit]):
    n = len(B)

    # Step 1
    for i in range(1, n):
        cx(A[i], B[i])

    # Step 2
    for i in range(n - 2, 0, -1):
        cx(A[i], A[i+1])

    # Step 3
    for i in range(0, n - 1):
        mcx([B[i], A[i]], A[i+1])

    # Step 4
    mcx([z, A[n-1]], B[n-1])

    # Step 5
    for i in range(n-2, -1, -1):
        mcx([B[i], A[i]], A[i+1])
        mcx([z, A[i]], B[i])

    # Step 6
    for i in range(1, n-1):
        cx(A[i], A[i+1])

    # Step 7
    for i in range(1, n):
        cx(A[i], B[i])

def control_add_sub(z: Qubit, A: list[Qubit], B: list[Qubit]):
    for i in A:
        cx(z, i)

    add(A, B)

    for i in A:
        cx(z, i)

def initial_subtraction(R : QuantumFloat, F : QuantumFloat, z : QuantumFloat):
    n = R.size

    # Step 1
    x(R[n-2])

    # Step 2
    cx(R[n-2], R[n-1])

    # Step 3
    cx(R[n-1], F[1])

    # Step 4
    mcx([R[n-1]], z[0], ctrl_state=0)
    
    # Step 5
    mcx([R[n-1]], F[2], ctrl_state=0)

    # Step 6
    control_add_sub(z, [R[n-4], R[n-3], R[n-2], R[n-1]], [F[0], F[1], F[2], F[3]])

def conditional_addition_or_subtraction(R : QuantumFloat, F : QuantumFloat, z : QuantumFloat):
    n = R.size

    for i in range(2, n // 2):
        # Step 1
        mcx(z, F[1], ctrl_state=0)

        # Step 2
        cx(F[2], z)

        # Step 3
        cx(R[n-1], F[1])

        # Step 4
        mcx([R[n-1]], z[0], ctrl_state=0)

        # Step 5
        mcx([R[n-1]], F[i+1], ctrl_state=0)

        # Step 6
        for j in range(i + 1, 2, -1):
            swap(F[j], F[j-1])

        # Step 7
        R_sum_qubits = [R[j] for j in range(n - 2 * i - 2, n)]
        F_sum_qubits = [F[j] for j in range(0, 2 * i + 2)]
        control_add_sub(z, R_sum_qubits, F_sum_qubits)

def remainder_restoration(R : QuantumFloat, F : QuantumFloat, z : QuantumFloat):
    n = R.size

    # Step 1
    mcx(z, F[1], ctrl_state=0)

    # Step 2
    cx(F[2], z)

    # Step 3
    mcx([R[n-1]], z[0], ctrl_state=0)

    # Step 4
    mcx([R[n-1]], F[n//2+1], ctrl_state=0)

    # Step 5
    x(z)

    # Step 6
    control_add(z, R, F)

    # Step 7
    x(z)

    # Step 8
    for j in range(n//2 + 1, 2, -1):
        swap(F[j], F[j-1])

    # Step 9
    cx(F[2], z)

def q_isqrt(R: QuantumFloat) -> QuantumFloat:
    """
    Computes the integer square root of a QuantumFloat, as well as the remainder using the `non-Restoring square root algorithm <https://arxiv.org/abs/1712.08254>`_.
    Does not yield plausible output when applied to negative integers.

    Parameters
    ----------
    R : QuantumFloat
        QuantumFloat value to compute the integer square root of.
    
    Returns
    -------
    QuantumFloat
        The integer square root of R.
    Also transforms input value R into the remainder.

    Examples
    --------
    Calculate the integer square root of 131:

    >>> from qrisp import QuantumFloat, q_isqrt
    >>> R = QuantumFloat(8, 0)
    >>> R[:] = 131
    >>> res = q_isqrt(R)
    >>> print(res.get_measurement())
    >>> print(R.get_measurement())
    {11: 1.0}
    {10: 1.0}
    """
    n = R.size
    e = R.exponent

    if e != 0:
        raise Exception("Tried to compute integer square root for QuantumFloat with non-zero exponent")
    
    if n == 1:
        F = QuantumFloat(1, 0)
        cx(R[0],F[0])
        cx(F[0],R[0])
        return F

    # To ensure that first qubit is 0 and total amount of qubits is even we extend the input variable
    delta = 1 if n % 2 == 1 else 2
    n += delta
    R.extend(delta)

    F = QuantumFloat(n, 0)
    z = QuantumFloat(1, 0)

    F[:] = 1
    z[:] = 0

    initial_subtraction(R, F, z)
    conditional_addition_or_subtraction(R, F, z)
    remainder_restoration(R, F, z)

    x(F[0])
    for i in range(2, n // 2 + 2):
        swap(F[i], F[i - 2])

    R.reduce(R[-delta:])
    F.reduce(F[-delta:])
    z.delete()

    return F


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""


def montgomery_decoder(y, R, N):
    if 0 < R < 1:
        R = modinv(R**-1, N)
    return (y * modinv(R, N)) % N


def montgomery_encoder(y, R, N):
    if 0 < R < 1:
        R = modinv(R**-1, N)
    return (int(y) % N * int(R) % N) % N


def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    else:
        g, y, x = egcd(b % a, a)
        return (g, x - (b // a) * y, y)


def modinv(a, m):
    g, x, y = egcd(a, m)
    if g != 1:
        raise Exception("modular inverse does not exist")
    else:
        return x % m



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np


def inverse_mod2(matrix):
    """
    Calculates the inverse of a binary matrix over the field of integers modulo 2.

    Parameters
    ----------
    matrix : numpy.array
        An invertible binary matrix.
    result : numpy.array
        The inverse of the given matrix over the field of integers modulo 2.

    """
    rows, columns = matrix.shape
    if rows != columns:
        raise Exception("The matrix must be a square matrix")

    matrix = matrix.copy()
    result = np.eye(rows, dtype=int)

    for i in range(rows):
        # Find the pivot row
        max_row = i + np.argmax(matrix[i:, i])

        # Swap the current row with the pivot row
        matrix[[i, max_row]] = matrix[[max_row, i]]
        result[[i, max_row]] = result[[max_row, i]]

        # Eliminate all rows below the pivot
        for j in range(i + 1, rows):
            if matrix[j, i] == 1:
                matrix[j] = (matrix[j] + matrix[i]) % 2
                result[j] = (result[j] + result[i]) % 2

    # Backward elimination
    for i in range(rows - 1, -1, -1):
        for j in range(i - 1, -1, -1):
            if matrix[j, i] == 1:
                matrix[j] = (matrix[j] + matrix[i]) % 2
                result[j] = (result[j] + result[i]) % 2

    return result


def gaussian_elimination_mod2(matrix, type="row", reduced=False, show_pivots=False):
    r"""
    Performs Gaussian elimination in the field $F_2$.

    Parameters
    ----------
    matrix : numpy.array
        A binary matrix.
    type : str, optional
        Available are ``row`` for row echelon form, and ``column`` for column echelon form.
        The default is ``row``.
    reduced : Boolean, optional
        If ``True``, the reduced row (column) echelon form is calculated.
        The default is ``False``.
    show_pivots : Boolean, optional
        If ``True``, the pivot columns (rows) are returned.

    Returns
    -------
    matrix : numpy.array
        The row (column) echelon form of the given matrix.
    pivots : list[int], optional
        A list of indices for the pivot columns (rows).

    """

    matrix = matrix.copy()

    rows, columns = matrix.shape
    column_index = 0
    row_index = 0
    pivots = []  # the pivot columns (type='row') or rows (type='column')

    if type == "row":
        while row_index < rows and column_index < columns:
            # Find the pivot row
            max_row = row_index + np.argmax(matrix[row_index:, column_index])

            if matrix[max_row, column_index] == 0:
                column_index += 1
            else:
                # Pivot
                pivots.append(column_index)

                # Swap the current row with the pivot row
                matrix[[row_index, max_row]] = matrix[[max_row, row_index]]

                # Eliminate all rows after the pivot
                for j in range(row_index + 1, rows):
                    if matrix[j, column_index] == 1:
                        matrix[j] = (matrix[j] + matrix[row_index]) % 2

                row_index += 1
                column_index += 1

        # Backward elimination (optional, for reduced row echelon form)
        if reduced:
            for row, col in enumerate(pivots):
                for i in range(row):
                    if matrix[i][col] == 1:
                        matrix[i] = (matrix[row] + matrix[i]) % 2

    elif type == "column":
        while row_index < rows and column_index < columns:
            # Find the pivot column
            max_column = column_index + np.argmax(matrix[row_index, column_index:])

            if matrix[row_index, max_column] == 0:
                row_index += 1
            else:
                # Pivot
                pivots.append(row_index)

                # Swap the current column with the pivot column
                matrix[:, [column_index, max_column]] = matrix[
                    :, [max_column, column_index]
                ]

                # Eliminate all columns after the pivot
                for j in range(column_index + 1, columns):
                    if matrix[row_index, j] == 1:
                        matrix[:, j] = (matrix[:, j] + matrix[:, column_index]) % 2

                row_index += 1
                column_index += 1

        # Backward elimination (optional, for reduced row echelon form)
        if reduced:
            for col, row in enumerate(pivots):
                for i in range(col):
                    if matrix[row][i] == 1:
                        matrix[:, i] = (matrix[:, col] + matrix[:, i]) % 2

    if show_pivots:
        return matrix, pivots
    else:
        return matrix


def construct_change_of_basis(S):
    """
    Implements the CZ construction outlined in https://quantum-journal.org/papers/q-2021-01-20-385/.

    Parameters
    ----------
    S : numpy.array
        A matrix representing a list of commuting Pauli operators: Each column is a Pauli operator in binary representation.

    Returns
    -------
    A : numpy.array
        Adjacency matrix for the graph state.
    R_inv : numpy.array
        A matrix representing a list of new Pauli operators.
    h_list : numpy.array
        A list indicating the qubits on which a Hadamard gate is applied.
    s_list : numpy.array
        A list indicating the qubits on which an S gate is applied.
    perm_vec : numpy.array
        A vector repesenting a permutation.

    """

    n = int(S.shape[0] / 2)

    ####################
    # Step 0: Calculate S_0: Independent columns (i.e., Pauli terms) of S
    ####################

    S_reduced, independent_cols = gaussian_elimination_mod2(
        S, reduced=True, show_pivots=True
    )
    k = len(independent_cols)

    S0 = S[:, independent_cols]
    R0_inv = S_reduced[:k, :]

    ####################
    # Step 1: Calculate S_1: The first k rows of the X component have full rank k
    ####################

    # Find independent rows in X component of S0
    S0X_reduced, independent_rows = gaussian_elimination_mod2(
        S0[-n:, :], type="column", show_pivots=True
    )

    S1 = S0.copy()
    h_list = []
    # Construct S1 by applying a Hadamard (i.e., a swap) to the rows of S0 not in independent_rows
    if len(independent_rows) < k:
        h_list = [i for i in range(n) if i not in independent_rows]
        for i in h_list:
            S1[[i, n + i]] = S1[[n + i, i]]

    # Find independent rows in X component of S1
    S1X_reduced, independent_rows = gaussian_elimination_mod2(
        S1[-n:, :], type="column", show_pivots=True
    )

    # Construct permutation achieving that the first k rows in X component of S1 are independent
    perm = np.arange(0, n)
    for index1, index2 in enumerate(independent_rows):
        curr = perm[index1]
        perm[index1] = index2
        perm[index2] = curr

    # Apply permutation to rows of S1 for Z and X component
    S1 = np.vstack((S1[:n, :][perm], S1[-n:, :][perm]))

    ####################
    # Step 2: Calculate S2: The first k rows of the X component are the identity matrix
    ####################

    R1_inv = S1[n : n + k, :]
    R1 = inverse_mod2(R1_inv)
    S2 = S1 @ R1 % 2

    ####################
    # Step 3: Calculate S3: Basis extension if n>k
    ####################

    if n > k:

        C = S2[:k, :]
        D = S2[k:n, :]
        F = S2[-(n - k) :, :]

        S3 = np.block(
            [
                [C, np.transpose(D)],
                [D, np.zeros((n - k, n - k), dtype=int)],
                [np.eye(k, dtype=int), np.zeros((k, n - k), dtype=int)],
                [F, np.eye(n - k, dtype=int)],
            ]
        )
        R2_inv = np.block([[np.eye(k, dtype=int)], [np.zeros((n - k, k), dtype=int)]])

    else:

        S3 = S2
        R2_inv = np.eye(n, dtype=int)

    ####################
    # Step 4: Calculate S4
    ####################

    R3_inv = S3[-n:, :]
    R3 = inverse_mod2(R3_inv)

    S4 = S3 @ R3 % 2

    # Remove diagonal entries in upper left block
    s_list = []
    for i in range(n):
        if S4[:n, :][i, i] == 1:
            s_list.append(i)

    Q2 = np.block(
        [
            [np.eye(n, dtype=int), S4[:n, :] * np.eye(n, dtype=int)],
            [np.zeros((n, n), dtype=int), np.eye(n, dtype=int)],
        ]
    )

    S4 = Q2 @ S4 % 2

    R_inv = R3_inv @ R2_inv @ R1_inv @ R0_inv % 2

    # Adjacency matrix for the graph
    A = S4[:n, :]

    return A, R_inv, h_list, s_list, perm



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax.numpy as jnp

from qrisp.jasp.interpreter_tools import (
    eval_jaxpr,
    extract_invalues,
    insert_outvalues,
    exec_eqn,
)


def evaluate_cond_eqn(cond_eqn, context_dic, eqn_evaluator=exec_eqn):

    # Extract the invalues from the context dic
    invalues = extract_invalues(cond_eqn, context_dic)

    for i in range(len(cond_eqn.params["branches"])):
        if int(invalues[0]) == i:
            res = eval_jaxpr(
                cond_eqn.params["branches"][i], eqn_evaluator=eqn_evaluator
            )(*invalues[1:])
            break

    if not isinstance(res, tuple):
        res = (res,)

    insert_outvalues(cond_eqn, context_dic, res)


def evaluate_while_loop(
    while_loop_eqn, context_dic, eqn_evaluator=exec_eqn, break_after_first_iter=False
):

    num_const_cond_args = while_loop_eqn.params["cond_nconsts"]
    num_const_body_args = while_loop_eqn.params["body_nconsts"]
    overall_constant_amount = max(num_const_cond_args, num_const_body_args)    

    def break_condition(invalues):
        
        constants = invalues[:num_const_cond_args]
        carries = invalues[overall_constant_amount:]
        
        new_invalues = constants + carries
        
        return eval_jaxpr(
            while_loop_eqn.params["cond_jaxpr"], eqn_evaluator=eqn_evaluator
        )(*new_invalues)

    # Extract the invalues from the context dic
    invalues = extract_invalues(while_loop_eqn, context_dic)
    outvalues = invalues[num_const_body_args:]

    while break_condition(invalues):
        
        constants = invalues[:num_const_body_args]
        carries = invalues[overall_constant_amount:]
        
        new_invalues = constants + carries

        outvalues = eval_jaxpr(
            while_loop_eqn.params["body_jaxpr"], eqn_evaluator=eqn_evaluator
        )(*new_invalues)

        # Update the non-const invalues
        
        if len(while_loop_eqn.params["body_jaxpr"].jaxpr.outvars) == 1:
            outvalues = (outvalues,)
        
        invalues = invalues[:overall_constant_amount] + list(outvalues)

        if break_after_first_iter:
            break

    insert_outvalues(while_loop_eqn, context_dic, outvalues)


def evaluate_scan(scan_eq, context_dic, eqn_evaluator=exec_eqn):

    invalues = extract_invalues(scan_eq, context_dic)

    f = eval_jaxpr(scan_eq.params["jaxpr"], eqn_evaluator=eqn_evaluator)

    length = scan_eq.params["length"]

    carry_amount = scan_eq.params["num_carry"]
    const_amount = scan_eq.params["num_consts"]

    init = invalues[const_amount : carry_amount + const_amount]

    if len(invalues) == carry_amount + const_amount:
        xs = [[]] * length
    else:
        xs = [[invalues[-1][i]] for i in range(length)]

    carry = init
    consts = invalues[:const_amount]
    ys = []
    for x in xs:
        args = consts + list(carry) + x

        res = f(*args)

        if not isinstance(res, tuple):
            res = (res,)

        carry = res[:carry_amount]
        y = res[carry_amount:]

        if len(y):
            ys.append(y[0])

    if len(ys):
        ys = [jnp.stack(ys)]

    outvalues = list(carry) + ys

    insert_outvalues(scan_eq, context_dic, outvalues)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.lax import fori_loop, while_loop, cond
from jax.extend.core import ClosedJaxpr
import jax

from qrisp.core import recursive_qv_search, recursive_qa_search
from qrisp.jasp.tracing_logic import TracingQuantumSession, check_for_tracing_mode, get_last_equation
from qrisp.jasp.primitives import AbstractQuantumCircuit


def q_while_loop(cond_fun, body_fun, init_val):
    """
    Jasp compatible version of
    `jax.lax.while_loop <https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.while_loop.html#jax.lax.while_loop>`_
    The parameters and semantics are the same as for the Jax version.

    In particular the following loop is performed

    ::

        def q_while_loop(cond_fun, body_fun, init_val):
            val = init_val
            while cond_fun(val):
                val = body_fun(val)
            return val

    Parameters
    ----------
    cond_fun : callable
        A function that evaluates the condition of the while loop. Must not
        contain any quantum operations.
    body_fun : callable
        A function describing the body of the loop.
    init_val : object
        An object to initialize the loop.

    Raises
    ------
    Exception
        Tried to modify quantum state during while condition evaluation.

    Returns
    -------
    val
        The result of ``body_fun`` after the last iteration.

    Examples
    --------

    We write a dynamic loop that collects measurement values of a quantum
    qubits into an accumulator. Note that the accumulator variable is a carry
    value implying the loop could not be implemented using :ref:`jrange`.


    ::

        from qrisp import *
        from qrisp.jasp import *

        @jaspify
        def main(k):

            qf = QuantumFloat(6)

            def body_fun(val):
                i, acc, qf = val
                x(qf[i])
                acc += measure(qf[i])
                i += 1
                return i, acc, qf

            def cond_fun(val):
                return val[0] < 5

            i, acc, qf = q_while_loop(cond_fun, body_fun, (0, 0, qf))

            return acc, measure(qf)

        print(main(6))
        # Yields
        # (Array(5, dtype=int64), Array(31., dtype=float64))

    """

    if not check_for_tracing_mode():
        return while_loop(cond_fun, body_fun, init_val)

    def new_cond_fun(val):
        temp_qc = qs.abs_qc
        res = cond_fun(val[0])
        if not qs.abs_qc is temp_qc:
            raise Exception(
                "Tried to modify quantum state during while condition evaluation"
            )
        return res

    def new_body_fun(val):
        qs.start_tracing(val[1])
        # The QuantumVariables from the arguments went through a flatten/unflattening cycle.
        # The unflattening creates a new QuantumVariable object, that is however not yet
        # registered in any QuantumSession. We register these in the current QuantumSession.
        for qv in recursive_qv_search(val[0]):
            qs.register_qv(qv, None)
        res = body_fun(val[0])
        abs_qc = qs.conclude_tracing()
        return (res, abs_qc)

    qs = TracingQuantumSession.get_instance()
    abs_qc = qs.abs_qc
    new_init_val = (init_val, abs_qc)
    while_res = while_loop(new_cond_fun, new_body_fun, new_init_val)

    eqn = get_last_equation()
    
    body_jaxpr = eqn.params["body_jaxpr"]
    
    # If the AbstractQuantumCircuit is part of the constants of the body,
    # the body did not execute any quantum operations.
    # We remove the AbstractQuantumCircuit from the body signature
    # to make the loop purely classical.
    for i in range(eqn.params["body_nconsts"]):
        if isinstance(body_jaxpr.jaxpr.invars[i].aval, AbstractQuantumCircuit):
            eqn.invars.pop(i)
            body_jaxpr.jaxpr.invars.pop(i)
            eqn.params["body_nconsts"] -= 1
            return while_res[0]
        
    from qrisp import Jaspr

    eqn.params["body_jaxpr"] = Jaspr.from_cache(body_jaxpr)

    qs.abs_qc = while_res[1]
    return while_res[0]


def q_fori_loop(lower, upper, body_fun, init_val):
    """
    Jasp compatible version of
    `jax.lax.fori_loop <https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.fori_loop.html#jax.lax.fori_loop>`_
    The parameters and semantics are the same as for the Jax version.

    In particular the following loop is performed

    ::

        def q_fori_loop(lower, upper, body_fun, init_val):
            val = init_val
            for i in range(lower, upper):
                val = body_fun(i, val)
            return val

    Parameters
    ----------
    lower : int or jax.core.Tracer
        An integer representing the loop index lower bound (inclusive).
    upper : int or jax.core.Tracer
        An integer representing the loop index upper bound (exclusive).
    body_fun : callable
        The function describing the loop body.
    init_val : object
        Some object to initialize the loop with.

    Returns
    -------
    val : object
        The return value of body_fun after the final iteration.

    Examples
    --------

    We write a dynamic loop that collects measurement values of a quantum
    qubits into an accumulator:

    ::

        @jaspify
        def main(k):

            qf = QuantumFloat(6)

            def body_fun(i, val):
                acc, qf = val
                x(qf[i])
                acc += measure(qf[i])
                return acc, qf

            acc, qf = q_fori_loop(0, k, body_fun, (0, qf))

            return acc, measure(qf)

        print(main(k))
        # Yields:
        # (Array(5, dtype=int64), Array(31., dtype=float64))

    """

    def new_body_fun(val):
        body_val = val[0]
        i = val[1]
        return (body_fun(i, body_val), i + 1, val[2])

    def new_cond_fun(val):
        i = val[1]
        upper = val[2]
        return i < upper

    return q_while_loop(new_cond_fun, new_body_fun, (init_val, lower, upper))[0]


def q_cond(pred, true_fun, false_fun, *operands):
    r"""
    Jasp compatible version of
    `jax.lax.cond <https://jax.readthedocs.io/en/latest/_autosummary/jax.lax.cond.html#jax.lax.cond>`_
    The parameters and semantics are the same as for the Jax version.

    Performs the following semantics:

    ::

        def q_cond(pred, true_fun, false_fun, *operands):
            if pred:
                return true_fun(*operands)
            else:
                return false_fun(*operands)


    Parameters
    ----------
    pred : bool or jax.core.Tracer
        A boolean value, deciding which function gets executed.
    true_fun : callable
        The function that is executed when ``pred`` is True.
    false_fun : callable
        The function that is executed when ``pred`` is False.
    *operands : tuple
        The input values for both functions.

    Returns
    -------
    object
        The return value of the respective function.

    Examples
    --------

    We write a script that brings a :ref:`QuantumBool` into superpostion and
    subsequently measures it. If the measurement result is ``False`` we flip
    it such that in the end, the bool will always be in the $\ket{\text{True}}$
    state.

    ::

        from qrisp import *
        from qrisp.jasp import *

        @jaspify
        def main():

            def false_fun(qbl):
                qbl.flip()
                return qbl

            def true_fun(qbl):
                return qbl

            qbl = QuantumBool()
            h(qbl)
            pred = measure(qbl)

            qbl = q_cond(pred,
                         true_fun,
                         false_fun,
                         qbl)

            return measure(qbl)

        print(main())
        # Yields:
        # True

    """

    def new_true_fun(*operands):
        qs.start_tracing(operands[1])
        for qv in recursive_qv_search(operands[0]):
            qs.register_qv(qv, None)
        res = true_fun(*operands[0])
        abs_qc = qs.conclude_tracing()
        return (res, abs_qc)

    def new_false_fun(*operands):
        qs.start_tracing(operands[1])
        for qv in recursive_qv_search(operands[0]):
            qs.register_qv(qv, None)
        res = false_fun(*operands[0])
        abs_qc = qs.conclude_tracing()
        return (res, abs_qc)

    qs = TracingQuantumSession.get_instance()
    abs_qc = qs.abs_qc

    new_operands = (operands, abs_qc)

    cond_res = cond(pred, new_true_fun, new_false_fun, *new_operands)

    # There seem to be situations, where Jax performs some automatic type
    # conversion after the cond call. This results in the cond equation
    # not being the most recent equation.
    # We therefore search for the last cond primitive.
    i = 1
    while True:
        eqn = get_last_equation(-i)
        if eqn.primitive.name == "cond":
            break
        i += 1
        
    false_jaxpr = eqn.params["branches"][0]
    true_jaxpr = eqn.params["branches"][1]

    if not isinstance(false_jaxpr.jaxpr.invars[-1].aval, AbstractQuantumCircuit):
        raise Exception(
            "Found implicit variable import in q_cond. Please make sure all used variables are part of the body signature."
        )

    from qrisp.jasp import Jaspr

    eqn.params["branches"] = (
            Jaspr.from_cache(false_jaxpr),
            Jaspr.from_cache(true_jaxpr)
    )

    qs.abs_qc = cond_res[-1]

    return cond_res[0]



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax
from jax.extend.core import ClosedJaxpr, JaxprEqn

from qrisp.environments import QuantumEnvironment, control
from qrisp.environments.custom_control_environment import custom_control
from qrisp.circuit import Operation
from qrisp.core.session_merging_tools import recursive_qs_search, merge
from qrisp.misc import get_depth_dic
from qrisp.jasp import check_for_tracing_mode, get_last_equation


class ConjugationEnvironment(QuantumEnvironment):
    r"""
    This :ref:`QuantumEnvironment<QuantumEnvironment>` can be used for perfoming conjugated operations.
    An arbitrary unitary :math:`U \in SU(2^n)` can be conjugated by another unitary
    :math:`V \in SU(2^n)`:

    .. math::

        \text{conj}(U,V) = V^\dagger U V


    This structure appears in many quantum algorithms such as `Grover <https://arxiv.org/abs/quant-ph/9605043>`_,
    `Quantum backtracking <https://arxiv.org/abs/1509.02374>`_ or
    `Fourier arithmetic <https://arxiv.org/abs/quant-ph/0008033>`_.

    Using the ``ConjugationEnvironment`` not only helps to structure the code,
    but can also grant performance advantages.

    This is because the controlled circuit of such a conjugation is can be
    realized by just controlling :math:`U` instead of all three operations.

    .. math::

        C\text{conj}(U,V) = V^\dagger CU V


    The ``ConjugationEnvironment`` can be called using the alias ``conjugate``.
    Conjugate takes the conjugation function (in our example :math:`V`) and returns
    a function that takes the arguments for the conjugation function and returns
    the corresponding ``ConjugationEnvironment``.
    For more information consult the examples section.

    .. note::

        Note that every QuantumVariable that is created by the conjugation
        function :math:`V` must be deleted/uncomputed before function conclusion.

    Parameters
    ----------

    conjugation_function : function
        The function performing the operation :math:`V`.
    args : iterable
        The arguments for the conjugation function.
    kwargs : dict
        The keyword arguments for the conjugation function.

    Examples
    --------

    We perform Fourier addition on a :ref:`QuantumFloat`

    ::

        from qrisp import conjugate, QuantumFloat, p, QFT

        def fourier_adder(qf, n):

            with conjugate(QFT)(qf):

                for i in range(qf.size):
                    p(n*np.pi*2**(i-qf.size+1), qf[i])

    >>> qf = QuantumFloat(5)
    >>> fourier_adder(qf, 3)
    >>> print(qf)
    {3: 1.0}
    >>> fourier_adder(qf, 2)
    {5: 1.0}

    Investigate the effects of a controlled addition:

    ::

        from qrisp import control

        ctrl = QuantumFloat(1)
        qf = QuantumFloat(5)

        with control(ctrl):
            fourier_adder(qf, 3)


    To see that indeed only the conjugand has been controlled we take a look
    at the circuit:

    >>> print(qf.qs.transpile(1))

    .. code-block:: none

        ctrl.0: ─────────■──────────■─────────■─────────■─────────■─────────────────
                ┌──────┐ │P(3π/16)  │         │         │         │      ┌─────────┐
          qf.0: ┤0     ├─■──────────┼─────────┼─────────┼─────────┼──────┤0        ├
                │      │            │P(3π/8)  │         │         │      │         │
          qf.1: ┤1     ├────────────■─────────┼─────────┼─────────┼──────┤1        ├
                │      │                      │P(3π/4)  │         │      │         │
          qf.2: ┤2 QFT ├──────────────────────■─────────┼─────────┼──────┤2 QFT_dg ├
                │      │                                │P(3π/2)  │      │         │
          qf.3: ┤3     ├────────────────────────────────■─────────┼──────┤3        ├
                │      │                                          │P(3π) │         │
          qf.4: ┤4     ├──────────────────────────────────────────■──────┤4        ├
                └──────┘                                                 └─────────┘


    """

    def __init__(self, conjugation_function, args, kwargs, allocation_management=True):

        self.conjugation_function = conjugation_function

        self.args = args

        self.kwargs = kwargs

        self.manual_allocation_management = allocation_management

        QuantumEnvironment.__init__(self)

    def __enter__(self):

        QuantumEnvironment.__enter__(self)

        if check_for_tracing_mode():
            with PJITEnvironment():
                res = self.conjugation_function(*list(self.args), **self.kwargs)
                # res = qache(self.conjugation_function)(*list(self.args), **self.kwargs)
            return res

        merge(recursive_qs_search(self.args) + [self.env_qs])

        qv_set_before = set(self.env_qs.qv_list)
        
        try:
            res = self.conjugation_function(*self.args, **self.kwargs)
        except Exception as e:
            QuantumEnvironment.__exit__(self, type(e), e, "")
            raise e

        temp_data = list(self.env_qs.data)
        self.env_qs.data = []
        i = 0

        while temp_data:
            instr = temp_data.pop(i)
            if isinstance(instr, QuantumEnvironment):
                instr.compile()
            else:
                self.env_qs.append(instr)

        creation_dic = {}
        for i in range(len(self.env_qs.data)):
            instr = self.env_qs.data[i]
            if instr.op.name == "qb_alloc":
                creation_dic[instr.qubits[0]] = 1
            elif instr.op.name == "qb_dealloc":
                if instr.qubits[0] not in creation_dic:
                    raise Exception(f"Tried to destroy qubit {instr.qubits[0]} within a conjugator.")
                else:
                    creation_dic[instr.qubits[0]] -=1
                    
        for k, v in creation_dic.items():
            if v != 0:
                raise Exception(f"Tried to create qubit {k} within a conjugator.")

        self.conjugation_circ = self.env_qs.copy()

        self.env_qs.data = []

        return res

    def __exit__(self, exception_type, exception_value, traceback):

        if exception_value:
            QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

        if not check_for_tracing_mode():
            conjugation_center_data = list(self.env_qs.data)
            self.env_qs.data = []
            self.perform_conjugation(conjugation_center_data)

        else:
            from qrisp.environments import invert

            with invert():
                with PJITEnvironment():
                    self.conjugation_function(*list(self.args), **self.kwargs)
                    # qache(self.conjugation_function)(*list(self.args), **self.kwargs)

        QuantumEnvironment.__exit__(self, exception_type, exception_value, traceback)

    @custom_control
    def perform_conjugation(self, conjugation_center_data, ctrl=None, ctrl_method=None):

        for instr in self.conjugation_circ.data:
            self.env_qs.append(instr)

        if ctrl is not None:
            with control(ctrl, ctrl_method=ctrl_method):
                self.env_qs.data.extend(conjugation_center_data)
        else:
            self.env_qs.data.extend(conjugation_center_data)

        for instr in self.conjugation_circ.inverse().data:
            self.env_qs.append(instr)

    def jcompile(self, eqn, context_dic):

        from qrisp.jasp import extract_invalues, insert_outvalues

        args = extract_invalues(eqn, context_dic)
        body_jaspr = eqn.params["jaspr"]

        flattened_jaspr = body_jaspr.flatten_environments()

        controlled_flattened_jaspr = flattened_jaspr.control(1)

        import jax

        def copy_jaxpr_eqn(jaxpr_eqn):
            return JaxprEqn(
                invars=list(jaxpr_eqn.invars),
                outvars=list(jaxpr_eqn.outvars),
                params=dict(jaxpr_eqn.params),
                primitive=jaxpr_eqn.primitive,
                effects=jaxpr_eqn.effects,
                source_info=jaxpr_eqn.source_info,
                ctx=jaxpr_eqn.ctx,
            )

        controlled_eqn_list = list(controlled_flattened_jaspr.eqns)
        controlled_eqn_list[0] = copy_jaxpr_eqn(controlled_flattened_jaspr.eqns[0])
        controlled_eqn_list[-1] = copy_jaxpr_eqn(controlled_flattened_jaspr.eqns[-1])

        controlled_eqn_list[0].invars.pop(0)
        controlled_eqn_list[-1].invars.pop(0)

        controlled_eqn_list[0].params["jaxpr"] = flattened_jaspr.eqns[0].params["jaxpr"]
        controlled_eqn_list[-1].params["jaxpr"] = flattened_jaspr.eqns[-1].params[
            "jaxpr"
        ]

        flattened_jaspr.ctrl_jaspr = controlled_flattened_jaspr.update_eqns(
            controlled_eqn_list
        )

        res = jax.jit(flattened_jaspr.eval)(*args)

        # Retrieve the equation
        jit_eqn = get_last_equation()
        
        jit_eqn.params["jaxpr"] = flattened_jaspr
        
        jit_eqn.params["name"] = "conjugation_env"

        if not isinstance(res, tuple):
            res = (res,)

        insert_outvalues(eqn, context_dic, res)

    def compile_(self, ctrl=None):

        temp = list(self.env_qs.data)
        self.env_qs.data = []

        for instr in self.conjugation_circ.data:
            if isinstance(instr, QuantumEnvironment):
                instr.compile()
            else:
                self.env_qs.append(instr)

        self.conjugation_circ = self.env_qs.copy()
        self.env_qs.data = []

        QuantumEnvironment.compile(self)

        content_circ = self.env_qs.copy()
        self.conjugation_circ.qubits = list(content_circ.qubits)

        conjugation_depth_dic = get_depth_dic(self.conjugation_circ)
        content_depth_dic = get_depth_dic(content_circ)

        added_depth_dic = {
            qb: conjugation_depth_dic[qb] + content_depth_dic[qb]
            for qb in content_circ.qubits
        }

        instruction_qubits = []

        i = 0
        while i < len(content_circ.qubits):

            qb = content_circ.qubits[i]

            if added_depth_dic[qb]:
                instruction_qubits.append(qb)
                i += 1
            else:
                content_circ.qubits.pop(i)
                self.conjugation_circ.qubits.pop(i)

        self.env_qs.data = temp

        conj_op = ConjugatedOperation(self.conjugation_circ, content_circ)

        alloc_instr = [
            instr
            for instr in self.conjugation_circ.data + content_circ.data
            if instr.op.name == "qb_alloc"
        ]

        for instr in alloc_instr:
            self.env_qs.append(instr)

        self.env_qs.append(conj_op, content_circ.qubits)

        dealloc_instr = [
            instr
            for instr in self.conjugation_circ.data + content_circ.data
            if instr.op.name == "qb_dealloc"
        ]

        for instr in dealloc_instr:
            self.env_qs.append(instr)


class ConjugatedOperation(Operation):

    def __init__(self, conjugation_circ, content_circ):

        self.conjugation_gate = conjugation_circ.to_gate(name="conjugator")
        self.content_gate = content_circ.to_gate(name="conjugand")

        definition = conjugation_circ.clearcopy()

        definition.append(self.conjugation_gate, definition.qubits)
        definition.append(self.content_gate, definition.qubits)
        definition.append(self.conjugation_gate.inverse(), definition.qubits)

        Operation.__init__(
            self,
            name="conjugation_env",
            definition=definition,
            num_qubits=definition.num_qubits(),
        )

    def control(self, num_ctrl_qubits=1, ctrl_state=-1, method=None):

        controlled_conjugand = self.content_gate.control(
            num_ctrl_qubits=num_ctrl_qubits, ctrl_state=ctrl_state, method=None
        )

        res = type(controlled_conjugand)(
            self, num_ctrl_qubits=num_ctrl_qubits, ctrl_state=ctrl_state, method=method
        )

        res.definition.data = []

        res.definition.append(self.conjugation_gate, self.definition.qubits)
        res.definition.append(
            controlled_conjugand,
            res.definition.qubits[:num_ctrl_qubits] + self.definition.qubits,
        )
        res.definition.append(self.conjugation_gate.inverse(), self.definition.qubits)

        return res

    def inverse(self):
        return ConjugatedOperation(
            self.conjugation_gate.definition, self.content_gate.inverse().definition
        )


def conjugate(conjugation_function, allocation_management=True):

    def conjugation_env_creator(*args, **kwargs):

        return ConjugationEnvironment(
            conjugation_function,
            args,
            kwargs,
            allocation_management=allocation_management,
        )

    return conjugation_env_creator


class PJITEnvironment(QuantumEnvironment):

    def jcompile(self, eqn, context_dic):

        from qrisp.jasp import extract_invalues, insert_outvalues, Jaspr

        args = extract_invalues(eqn, context_dic)
        body_jaspr = eqn.params["jaspr"]

        flattened_jaspr = body_jaspr.flatten_environments()

        res = jax.jit(flattened_jaspr.eval)(*args)

        jit_eqn = get_last_equation()
        
        jit_eqn.params["jaxpr"] = Jaspr.from_cache(jit_eqn.params["jaxpr"])

        if not isinstance(res, tuple):
            res = (res,)

        insert_outvalues(eqn, context_dic, res)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.jasp.primitives import get_qubit, slice_qb_array, get_size, fuse_qb_array, AbstractQubit


class DynamicQubitArray:

    def __init__(self, tracer):
        self.tracer = tracer

    def __getitem__(self, key):
        tracer = self.tracer
        if isinstance(key, slice):
            start = key.start
            if key.start is None:
                start = 0
            stop = key.stop
            if key.stop is None:
                stop = get_size(tracer)

            return DynamicQubitArray(slice_qb_array(tracer, start, stop))
        else:
            from qrisp.jasp.tracing_logic.tracing_quantum_session import (
                TracingQuantumSession,
            )

            qs = TracingQuantumSession.get_instance()
            id_tuple = (id(tracer), id(key))
            if not id_tuple in qs.qubit_cache:
                qs.qubit_cache[id_tuple] = get_qubit(tracer, key)
            return qs.qubit_cache[id_tuple]

    @property
    def size(self):
        return get_size(self.tracer)

    def __add__(self, other):
        if isinstance(other, DynamicQubitArray):
            other = other.tracer
        if isinstance(other, list):
            temp = self
            for x in other:
                if not isinstance(other, AbstractQubit):
                    raise Exception("Can only concatenate type AbstractQubit or list[AbstractQubit] to DynamicQubitArray")
                temp = temp + x
            return temp
        return DynamicQubitArray(fuse_qb_array(self.tracer, other))
    
    def __radd__(self, other):
        if isinstance(other, DynamicQubitArray):
            other = other.tracer
        if isinstance(other, list):
            temp = self
            for x in other[::-1]:
                if not isinstance(other, AbstractQubit):
                    raise Exception("Can only concatenate type AbstractQubit or list[AbstractQubit] to DynamicQubitArray")
                temp = x + temp
            return temp
        
        return DynamicQubitArray(fuse_qb_array(other, self.tracer))
        

    @property
    def reg(self):
        return self

    def measure(self):
        from qrisp.jasp import TracingQuantumSession, Measurement_p

        qs = TracingQuantumSession.get_instance()
        res, qs.abs_qc = Measurement_p.bind(self.tracer, qs.abs_qc)
        return res


from jax import tree_util
from builtins import id


def flatten_dqa(dqa):
    return (dqa.tracer,), None


def unflatten_dqa(aux_data, children):
    return DynamicQubitArray(children[0])


tree_util.register_pytree_node(DynamicQubitArray, flatten_dqa, unflatten_dqa)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp.alg_primitives.mcx_algs.circuit_library import (
    gidney_qc,
    gidney_qc_inv,
    margolus_qc,
    ctrl_state_wrap,
)
from qrisp.circuit import QuantumCircuit, Operation


class GidneyLogicalAND(Operation):

    def __init__(self, inv=False, ctrl_state="11"):

        definition = QuantumCircuit(3)

        if ctrl_state[0] == "0":
            definition.x(definition.qubits[0])
        if ctrl_state[1] == "0":
            definition.x(definition.qubits[1])

        definition.append(margolus_qc.to_gate("margolus"), definition.qubits)

        if ctrl_state[0] == "0":
            definition.x(definition.qubits[0])
        if ctrl_state[1] == "0":
            definition.x(definition.qubits[1])

        name = "uncompiled_gidney_mcx"
        if inv:
            name = name + "_inv"
        Operation.__init__(self, name=name, num_qubits=3, definition=definition)

        self.permeability = {0: True, 1: True, 2: False}
        self.inv = inv
        self.is_qfree = True
        self.ctrl_state = ctrl_state

    def inverse(self):
        return GidneyLogicalAND(inv=not self.inv, ctrl_state=self.ctrl_state)
        # Implementation from https://arxiv.org/abs/2101.04764

    def recompile(self, computation_strategy="gidney"):
        if self.inv:
            compiled_qc = gidney_qc_inv
        else:
            if computation_strategy == "gidney":
                compiled_qc = gidney_qc
            elif computation_strategy == "margolus":
                compiled_qc = margolus_qc
            else:
                raise Exception(
                    f"Don't know measurement based uncomputation strategy {computation_strategy}"
                )

        name = "compiled_gidney_mcx"
        if self.inv:
            name = name + "_inv"

        res = ctrl_state_wrap(compiled_qc, self.ctrl_state).to_op(name)
        res.is_qfree = True
        res.permeability = {0: True, 1: True, 2: False}
        res.ctrl_state = self.ctrl_state
        return res


from qrisp.jasp import AbstractQubit, make_jaspr, Jaspr
from qrisp.core import x, h, cx, t, t_dg, s, measure, cz
from qrisp.environments import control


def gidney_mcx_impl(a, b, c):

    h(c)
    t(c)

    cx(a, c)
    cx(b, c)
    cx(c, a)
    cx(c, b)

    t_dg(a)
    t_dg(b)
    t(c)

    cx(c, a)
    cx(c, b)

    h(c)
    s(c)


def gidney_mcx_inv_impl(a, b, c):
    h(c)
    bl = measure(c)

    with control(bl):
        cz(a, b)
        x(c)


class GidneyMCXJaspr(Jaspr):

    slots = ["inv"]

    def __init__(self, inv):
        self.inv = inv
        if self.inv:
            temp_jaspr = make_jaspr(gidney_mcx_inv_impl)(
                AbstractQubit(), AbstractQubit(), AbstractQubit()
            ).flatten_environments()
        else:
            temp_jaspr = make_jaspr(gidney_mcx_impl)(
                AbstractQubit(), AbstractQubit(), AbstractQubit()
            ).flatten_environments()

        Jaspr.__init__(self, temp_jaspr)
        self.envs_flattened = True

    def inverse(self):
        if self.inv:
            return gidney_mcx_jaspr
        else:
            return gidney_mcx_inv_jaspr


gidney_mcx_jaspr = GidneyMCXJaspr(False)
gidney_mcx_inv_jaspr = GidneyMCXJaspr(True)


def jasp_gidney_mcx(a, b, c):
    gidney_mcx_jaspr.embedd(a, b, c, name="gidney_mcx")


def jasp_gidney_mcx_inv(a, b, c):
    gidney_mcx_inv_jaspr.embedd(a, b, c, name="gidney_mcx_inv")



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from qrisp import QuantumVariable, QuantumBool, x, mcx
from qrisp.algorithms.qaoa.mixers import controlled_RX_mixer_gen


def create_min_set_cover_mixer(sets, universe):
    r"""
    Creates the ``controlled_RX_mixer`` for an instance of the minimum set cover problem following `Hadfield et al. <https://arxiv.org/abs/1709.03489>`_

    The belonging ``predicate`` function indicates if a set can be swapped out of the solution.

    Parameters
    ----------
    sets : list[set]
        A list of sets.
    universe : set
        The union of all sets.

    Returns
    -------
    function
        A Python function receiving a :ref:`QuantumVariable` and real parameter $\beta$.
        This function performs the application of the mixer associated to the problem instance.

    """

    membership_dict = {
        element: [i for i, subset in enumerate(sets) if element in subset]
        for element in universe
    }

    def predicate(qv, i):
        anc = QuantumVariable(len(sets[i]))
        x(anc)
        for anc_index, element in enumerate(sets[i]):
            other_sets = [item for item in membership_dict[element] if item != i]
            mcx(
                [qv[set_index] for set_index in other_sets],
                anc[anc_index],
                ctrl_state="0" * len(other_sets),
            )
        qbl = QuantumBool()
        mcx(anc, qbl)
        return qbl

    controlled_RX_mixer = controlled_RX_mixer_gen(predicate)

    return controlled_RX_mixer


def create_min_set_cover_cl_cost_function(sets, universe):
    """
    Creates the classical cost function for an instance of the minimum set cover problem.

    Parameters
    ----------
    sets : list[set]
        A list of sets.
    universe : set
        The union of all sets.

    Returns
    -------
    cl_cost_function : function
        The classical cost function for the problem instance, which takes a dictionary of measurement results as input.

    """

    def cl_cost_function(res_dic):
        cost = 0
        for state, prob in res_dic.items():
            indices = [index for index, value in enumerate(state) if value == "1"]
            solution_sets = [sets[index] for index in indices]
            if len(solution_sets) > 0 and set.union(*solution_sets) == universe:
                cost += len(indices) * prob
            else:
                cost += len(sets)

        return cost

    return cl_cost_function


def min_set_cover_init_function(qv):
    r"""
    Prepares the initial state $\ket{1}^{\otimes n}$.

    Parameters
    ----------
    qv : :ref:`QuantumVariable`
        The quantum argument.

    """
    x(qv)


def min_set_cover_problem(sets, universe):
    """
    Creates a QAOA problem instance with appropriate phase separator, mixer, and
    classical cost function.

    Parameters
    ----------
    sets : list[set]
        A list of sets.
    universe : set
        The union of all sets.

    Returns
    -------
    :ref:`QAOAProblem`
        A QAOA problem instance for MinSetCover for given ``sets`` and ``universe``.

    """
    from qrisp.qaoa import QAOAProblem, RZ_mixer

    return QAOAProblem(
        cost_operator=RZ_mixer,
        mixer=create_min_set_cover_mixer(sets, universe),
        cl_cost_function=create_min_set_cover_cl_cost_function(sets, universe),
        init_function=min_set_cover_init_function,
    )



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import jax
import jax.numpy as jnp
from jax.scipy.optimize import OptimizeResults

# https://link.springer.com/chapter/10.1007/978-94-015-8330-5_4
# Constraints not included in this implementation


def cobyla(fun, x0, args, maxiter=50, cons=[], rhobeg=1.0, rhoend=1e-6, seed=3):
    r"""

    Minimize a scalar function of one or more variables using the Constrained Optimization By Linear Approximation (COBYLA) algorithm.

    Parameters
    ----------
        maxiter : int
            Maximum number of iterations to perform. Each iteration requires several function evaluations.
        rhobeg : float
            Initial simplex size. Determines how far from the initial guess the algorithm first explores.
        rhoend : float
            Ending simplex size. When the simplex contracts to around this scale, the algorithm terminates.

    Returns
    -------
    results
        An `OptimizeResults <https://docs.jax.dev/en/latest/_autosummary/jax.scipy.optimize.OptimizeResults.html#jax.scipy.optimize.OptimizeResults>`_ object.

    """

    def arg_fun(x):
        return fun(x, *args)

    n = len(x0)
    sim = jnp.zeros((n + 1, n))
    sim = sim.at[0].set(x0)
    sim = sim.at[1:].set(x0 + jnp.eye(n) * rhobeg)

    # Initialize function values and constraint values
    f = jax.lax.map(arg_fun, sim)
    c = jax.vmap(lambda x: jnp.array([con(x) for con in cons]))(sim)

    def body_fun(state):
        sim, f, c, rho, nfeval, niter = state

        # Find the best and worst points
        best = jnp.argmin(f)
        worst = jnp.argmax(f)

        # Calculate the centroid of the simplex excluding the worst point
        mask = jnp.arange(n + 1) != worst
        centroid = jnp.sum(sim * mask[:, None], axis=0) / n

        # Reflect the worst point
        xr = 2 * centroid - sim[worst]
        fr = arg_fun(xr)
        cr = jnp.array([con(xr) for con in cons])
        nfeval += 1

        # Expansion
        xe = 2 * xr - centroid
        fe = arg_fun(xe)
        ce = jnp.array([con(xe) for con in cons])
        nfeval += 1

        # Contraction
        xc = 0.5 * (centroid + sim[worst])
        fc = arg_fun(xc)
        cc = jnp.array([con(xc) for con in cons])
        nfeval += 1

        # Update simplex based on conditions
        cond_reflect = (fr < f[best]) & jnp.all(cr >= 0)
        cond_expand = (fe < fr) & cond_reflect
        cond_contract = (fc < f[worst]) & jnp.all(cc >= 0)

        sim = jnp.where(
            cond_expand,
            sim.at[worst].set(xe),
            jnp.where(
                cond_reflect,
                sim.at[worst].set(xr),
                jnp.where(
                    cond_contract, sim.at[worst].set(xc), 0.5 * (sim + sim[best])
                ),
            ),
        )

        f = jax.lax.map(arg_fun, sim)
        c = jax.vmap(lambda x: jnp.array([con(x) for con in cons]))(sim)
        nfeval += n + 1

        rho *= 0.5
        return sim, f, c, rho, nfeval, niter + 1

    def cond_fun(state):
        _, _, _, rho, nfeval, niter = state
        return (rho > rhoend) & (niter < maxiter)

    from qrisp.jasp import make_tracer

    state = (sim, f, c, rhobeg, make_tracer(n + 1), make_tracer(0))
    sim, f, _, _, nfeval, niter = jax.lax.while_loop(cond_fun, body_fun, state)
    best = jnp.argmin(f)
    x = sim[best]
    fx = arg_fun(x)

    return OptimizeResults(x, True, 0, fx, None, None, nfeval + 1, 0, niter)



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

from jax.extend.core import ClosedJaxpr
from jax import jit
from qrisp.jasp.interpreter_tools import (
    eval_jaxpr,
    extract_invalues,
    insert_outvalues,
    reinterpret,
)


def evaluate_pjit_eqn(pjit_eqn, context_dic):

    definition_jaxpr = pjit_eqn.params["jaxpr"].jaxpr

    # Extract the invalues from the context dic
    invalues = extract_invalues(pjit_eqn, context_dic)

    res = jit(eval_jaxpr(definition_jaxpr), inline=True)(*invalues)

    if len(definition_jaxpr.outvars) == 1:
        res = [res]

    # Insert the values into the context_dic
    insert_outvalues(pjit_eqn, context_dic, res)


# Flattens/Inlines a pjit calls in a jaxpr
def flatten_pjit(jaxpr):

    def eqn_evaluator(eqn, context_dic):
        if eqn.primitive.name == "pjit":
            evaluate_pjit_eqn(eqn, context_dic)
        else:
            return True

    return type(jaxpr)(reinterpret(jaxpr, eqn_evaluator))



"""
\********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************/
"""

from qrisp.core.gate_application_functions import x, cx, mcx
from qrisp.environments import invert, control
from qrisp.jasp import jlen, jrange, while_loop, jnp


def compute_ladder_iterations(N):
    """
    Computes the number of iterations required to reduce a given size `N`
    to a value less than or equal to 2 by repeatedly halving it.
    This function uses a loop to simulate the halving process and counts
    the number of iterations needed.
    Args:
        N (int): The initial size to be reduced.
    Returns:
        int: The number of iterations required to reduce `N` to a value
        less than or equal to 2.
    """

    iterations = 1
    current_size = N

    def body_fun(val):
        i, current_size = val
        current_size = current_size // 2
        i += 1
        return i, current_size

    def cond_fun(val):
        return val[1] > 2

    i, current_size = while_loop(cond_fun, body_fun, (iterations, current_size))
    return i


def X_prime_func(i, k, n):
    """
    Simulate the lists produced by the recursive calls
    of the algorithm by returning the corresponding index.
    This function is derived from observing that the original algorithm generates
    qubit lists by the following pattern: Starting with a range of integers from 0 to N-1,
    skip the first index, select every other index, and append the second-to-last index
    if the list size is even.

    Example:
    N=25
    Iteration 1: [1, 3, 5, 7, 9, 11, 13, 15, 17, 19, 21, 23]
    Iteration 2: [3, 7, 11, 15, 19, 21]
    Iteration 3: [7, 15, 19]
    Iteration 4: [15]
    """
    return jnp.int64(
        (2 ** (k + 1) - 1)
        + 2 ** (k + 1) * i
        - 2**k * ((i + 1) == n // 2) * ((n + 1) & 1)
    )


def ladder1_synth_jax(qf):
    """
    Iteratively synthesizes a logarithmic-depth CNOT circuit equivalent to a CNOT ladder of length n-1.
    This construction is based on the recursive algorithm described in: https://arxiv.org/abs/2501.16802, Algorithm 1.
    Args:
        qf: QuantumFloat
            The quantum float representing the qubits to be used in the circuit.

    This implementation differs from the one in the paper in two key aspects:
        - It uses an iterative structure instead of a recursive one.
        - It leverages the 'X_prime_func' function to track qubit indices instead of relying on Python lists.
    These modifications are necessary because dynamic programming does not support recursive structures or dynamically sized lists.
    To address these limitations, we first calculate the number of recursive calls that would have been made in the original algorithm.
    Then, we use 'X_prime_func', which encapsulates the structure of the lists created during recursive calls, to reference qubits in the iterative process.
    Additionally, we avoid variable updates between iterations by using bit shifts, noting that N//2 is equivalent to N>>1.
    """
    N = jlen(qf)

    # with control(N == 1):
    #     raise ValueError("The number of qubits should be greater than or equal to 2.")

    # Add the first layer of gates
    for i in jrange(1, (N + 1) // 2 - 1):
        cx(qf[2 * i - 1], qf[2 * i])
    cx(qf[-2], qf[-1])

    # Compute the number of recursive call in the original algorithm
    max_iterations = compute_ladder_iterations(N)

    # Fill in the left part of the circuit
    for k in jrange(max_iterations - 2):
        # CL
        cx(
            qf[X_prime_func((N >> (k + 1)) - 2, k, N >> k)],
            qf[X_prime_func((N >> (k + 1)) - 1, k, N >> k)],
        )

        for i in jrange(1, ((N >> (k + 1)) + 1) // 2 - 1):
            cx(
                qf[X_prime_func(2 * i - 1, k, N >> k)],
                qf[X_prime_func(2 * i, k, N >> k)],
            )

    # Conditionally append the deepest layer
    with control(N >> (max_iterations - 1) == 2):
        cx(
            qf[X_prime_func(0, max_iterations - 2, N >> (max_iterations - 2))],
            qf[X_prime_func(1, max_iterations - 2, N >> (max_iterations - 2))],
        )

    # Fill in the right part of the circuit
    with invert():
        for k in jrange(max_iterations - 2):
            # CR
            cx(qf[X_prime_func(0, k, N >> k)], qf[X_prime_func(1, k, N >> k)])

            for i in jrange(1, ((N >> (k + 1)) + 1) // 2 - 1):
                cx(
                    qf[X_prime_func(2 * i, k, N >> k)],
                    qf[X_prime_func(2 * i + 1, k, N >> k)],
                )

    # Add the last layer of gates
    cx(qf[0], qf[1])
    for i in jrange(1, (N + 1) // 2 - 1):
        cx(qf[2 * i], qf[2 * i + 1])


def ladder2_synth_jax(x, y, method="khattar"):
    """
    Iteratively synthesizes a polylogarithmic-depth MCX circuit equivalent to a Toffoli ladder of length n-1.
    This construction is based on the recursive algorithm described in `Algorithm 2 <https://arxiv.org/abs/2501.16802>`__.
    It generalizes the ladder1_synth_jax by alternating x and y qubits. If the circutit generated
    by ladder1_synth_jax(n) contains cx(x[i], x[j]) then the circuit generated by ladder2_synth_jax(n) contains mcx(x[i] + y[i:j], x[j]).
    Every MCX gate uses the Khattar method for synthesis.

    Args:
        x: QuantumFloat
            The first set of qubits.
        y: QuantumFloat
            The second set of qubits.
        method: str
            The method to use for synthesizing the MCX gates. Default is 'khattar'.
    """

    N = jlen(x)

    for i in jrange(1, (N + 1) // 2 - 1):
        mcx(
            x[2 * i - 1] + y[2 * i - 1],
            x[2 * i],
            method=method,
        )
    mcx(x[-2] + y[-1], x[-1], method=method) #CHANGED

    max_iterations = compute_ladder_iterations(N)
    for k in jrange(max_iterations - 2):
        # CL
        mcx(
            x[X_prime_func((N >> (k + 1)) - 2, k, N >> k)]
            + y[
                X_prime_func((N >> (k + 1)) - 2, k, N >> k) : X_prime_func(
                    (N >> (k + 1)) - 1, k, N >> k
                )
            ],
            x[X_prime_func((N >> (k + 1)) - 1, k, N >> k)],
            method=method,
        )

        for i in jrange(1, ((N >> (k + 1)) + 1) // 2 - 1):

            mcx(
                x[X_prime_func(2 * i - 1, k, N >> k)]
                + y.reg[
                    X_prime_func(2 * i - 1, k, N >> k) : X_prime_func(2 * i, k, N >> k)
                ],
                x[X_prime_func(2 * i, k, N >> k)],
                method=method,
            )

    with control(N >> (max_iterations - 1) == 2):
        mcx(
            x[X_prime_func(0, max_iterations - 2, N >> (max_iterations - 2))]
            + y[
                X_prime_func(
                    0, max_iterations - 2, N >> (max_iterations - 2)
                ) : X_prime_func(1, max_iterations - 2, N >> (max_iterations - 2))
            ],
            x[X_prime_func(1, max_iterations - 2, N >> (max_iterations - 2))],
            method=method,
        )

    with invert():
        for k in jrange(max_iterations - 2):
            # CR
            mcx(
                x[X_prime_func(0, k, N >> k)]
                + y[X_prime_func(0, k, N >> k) : X_prime_func(1, k, N >> k)],
                x[X_prime_func(1, k, N >> k)],
                method=method,
            )

            for i in jrange(1, ((N >> (k + 1)) + 1) // 2 - 1):
                mcx(
                    x[X_prime_func(2 * i, k, N >> k)]
                    + y.reg[
                        X_prime_func(2 * i, k, N >> k) : X_prime_func(
                            2 * i + 1, k, N >> k
                        )
                    ],
                    x[X_prime_func(2 * i + 1, k, N >> k)],
                    method=method,
                )

    mcx([x[0], y[0]], x[1], method=method)
    for i in jrange(1, (N + 1) // 2 - 1):
        mcx([x[2 * i], y[2 * i]], x[2 * i + 1], method=method)


def remaud_adder(a, b, z):
    
    """
    In-place adder function based on `this paper, Algorithm 3 <https://arxiv.org/abs/2501.16802>`__.
    Performs the addition:

    ::

        b += a 

    with a polylogarithmic depth and no use of auxiliary qubits. 
    Both `a` and `b` should have the same size `N` and the result of the overflow is stored in the carry Qubit.

    Parameters
    ----------
    a : int or QuantumVariable or list[Qubit]
        The value that should be added.
    b : QuantumVariable or list[Qubit]
        The value that should be modified in the in-place addition.
    z : Qubit
        The carry value resulting from the overflow of the addition.

    Examples
    --------

    We add two integers:

    >>> from qrisp import QuantumFloat, remaud_adder
    >>> a = QuantumFloat(4)
    >>> b = QuantumFloat(4)
    >>> z = QuantumFloat(1)
    >>> a[:] = 4
    >>> b[:] = 5
    >>> remaud_adder(a,b,z)
    >>> print(b)
    {9: 1.0}

    """
    
    n = jlen(a)

    for i in jrange(1, n):
        cx(a[i], b[i])
    
    ladder1_synth_jax(a[1:] + z[:])

    with invert():
        ladder2_synth_jax(a[:] + z[:], b[:]) 
    
    for i in jrange(1, n):
        cx(a[i], b[i])

    x(b[1:-1])
    
    ladder2_synth_jax(a[:], b[:-1])
    
    x(b[1:-1])
    
    with invert():
        ladder1_synth_jax(a[1:])
    
    for i in jrange(n):
        cx(a[i], b[i])


"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import sympy
import jax

import qrisp.circuit.standard_operations as std_ops
from qrisp.jasp import check_for_tracing_mode, DynamicQubitArray, jlen


def append_operation(operation, qubits=[], clbits=[], param_tracers=[]):
    from qrisp import find_qs

    qs = find_qs(qubits)
    qs.append(operation, qubits, clbits, param_tracers=param_tracers)


def cx(control, target):
    """
    Applies a CX gate.

    Parameters
    ----------
    control : Qubit or list[Qubit] or QuantumVariable
        The Qubit to control on.
    target : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the X gate on.
    """
    append_operation(std_ops.CXGate(), [control, target])

    # std_ops.CXGate().append([qubits_0, qubits_1])

    return control, target


def cy(control, target):
    """
    Applies a CY gate.

    Parameters
    ----------
    control : Qubit or list[Qubit] or QuantumVariable
        The Qubit to control on.
    target : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Y gate on.
    """

    append_operation(std_ops.CYGate(), [control, target])
    return control, target


def cz(control, target):
    """
    Applies a CZ gate.

    Parameters
    ----------
    control : Qubit or list[Qubit] or QuantumVariable
        The Qubit to control on.
    target : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Z gate on.
    """

    append_operation(std_ops.CZGate(), [control, target])
    return control, target


def h(qubits):
    """
    Applies an H gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the H gate on.
    """
    append_operation(std_ops.HGate(), [qubits])

    return qubits


def x(qubits):
    """
    Applies an X gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the X gate on.
    """

    append_operation(std_ops.XGate(), [qubits])

    return qubits


def y(qubits):
    """
    Applies a Y gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Y gate on.
    """

    append_operation(std_ops.YGate(), [qubits])

    return qubits


def z(qubits):
    """
    Applies a Z gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the Z gate on.
    """
    append_operation(std_ops.ZGate(), [qubits])

    return qubits


def mcx(controls, target, method="auto", ctrl_state=-1, num_ancilla=1):
    r"""
    .. _mcx:

    Applies a multi-controlled X gate.

    The following methods are available:


    .. list-table::
        :widths: 20 80
        :header-rows: 1

        *   - Method
            - Description
        *   - ``gray``
            - Performs a gray code traversal which requires no ancillae but is rather inefficient for large numbers of control qubits.
        *   - ``gray_pt``/``gray_pt_inv``
            - More efficient but introduce extra phases that need to be uncomputed by performing the inverse of this gate on the same inputs. For more information on phase tolerance, check `this paper <https://iopscience.iop.org/article/10.1088/2058-9565/acaf9d/meta>`__.
        *   - ``balauca``
            - Method based on this `paper <https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf>`__ with logarithmic depth but requires many ancilla qubits.
        *   - ``khattar``
            - Method based on this `paper <https://arxiv.org/abs/2407.17966>`_, implements the n-controlled x with 2n − 3 Toffoli and linear depth using 1 clean ancilla. The case n=3 implements this `paper <https://arxiv.org/abs/2106.11513>`__.
        *   - ``maslov``
            - Documented `here <https://arxiv.org/abs/1508.03273>`_, requires less ancilla qubits but is only available for 4 or less control qubits.
        *   - ``yong``
            - Can be found int this `article <https://link.springer.com/article/10.1007/s10773-017-3389-4>`__.This method requires only a single ancilla and has moderate scaling in depth and gate count.
        *   - ``amy``
            - A Toffoli-circuit (ie. only two control qubits are possible), which (temporarily) requires one ancilla qubit. However, instead of the no-ancilla T-depth 4, this circuit achieves a T-depth of 2. Find the implementation details in `this paper <https://arxiv.org/pdf/1206.0758.pdf>`__.
        *   - ``jones``
            - Similar to ``amy`` but uses two ancilla qubits, and has a T-depth of 1. Read about it `here <https://arxiv.org/abs/1212.5069>`__.
        *   - ``gidney``
            - A very unique way for synthesizing a logical AND. The Gidney Logical AND performs a circuit with T-depth 1 to compute the truth value and performs another circuit involving a measurement and a classically controlled CZ gate for uncomputation. The uncomputation circuit has T-depth 0, such that the combined T-depth is 1. Requires no ancillae. More details `here <https://arxiv.org/abs/1709.06648>`__. Works only for two control qubits.
        *   - ``hybrid``
            - A flexible method which combines the other available methods, such that the amount of used ancillae is customizable. After several ``balauca``-layers, the recursion is canceled by either a ``yong``, ``maslov`` or ``gray`` mcx, depending on what fits the most.
        *   - ``auto``
            - Recompiles the mcx gate at compile time using the hybrid algorithm together with the information about how many clean/dirty ancillae qubits are available. For more information check :meth:`qrisp.QuantumSession.compile`.

    .. note::
        Due to Qrisp's automatic qubit management, clean ancilla qubits are not as much
        of a sparse resource as one might think. Even though the ``balauca`` method
        requires a considerable amount of ancillae, many other functions also do,
        implying there is alot of recycling potential. The net effect is that in more
        complex programs, the amount of qubits of the circuit returned by the
        :meth:`compile method <qrisp.QuantumSession.compile>` increases only slightly.


    Parameters
    ----------
    controls : list[Qubits] or QuantumVariable
        The Qubits to control on.
    target : Qubit or QuantumBool
        The Qubit to perform the X gate on.
    method : str, optional
        The synthesis method. Available are ``auto``, ``gray``, ``gray_pt``,
        ``gray_pt_inv``, ``maslov``, ``balauca`` and ``yong``. The default is ``auto``.
    ctrl_state : int or str, optional
        The state on which to activate the X gate. The default is "1111..".
    num_ancilla : int, optional
        Specifies the amount of ancilla qubits to use. This parameter is used only if
        the method is set to ``hybrid``. The default is 1.


    Examples
    --------

    We apply a 3-contolled X gate

    >>> from qrisp import QuantumVariable, mcx, QuantumBool
    >>> control = QuantumVariable(3)
    >>> target = QuantumBool()
    >>> mcx(control, target, method = "gray")
    >>> print(control.qs)

    .. code-block:: none

        QuantumCircuit:
        --------------
        control.0: ──■──
                     │
        control.1: ──■──
                     │
        control.2: ──■──
                   ┌─┴─┐
         target.0: ┤ X ├
                   └───┘
        Live QuantumVariables:
        ---------------------
        QuantumVariable control
        QuantumBool target

    We compare different performance indicators. ::

        from qrisp import QuantumVariable, mcx, QuantumBool

        def benchmark_mcx(n, methods):
            for method in methods:

                controls = QuantumVariable(n)
                target = QuantumBool()

                mcx(controls, target, method = method)

                compiled_qc = controls.qs.compile()

                print(f"==================\nMethod: {method}\n------------------")
                print(f"Depth: \t\t\t{compiled_qc.depth()}")
                print(f"CNOT count: \t{compiled_qc.cnot_count()}")
                print(f"Qubit count: \t{compiled_qc.num_qubits()}")


    >>> benchmark_mcx(4, methods = ["gray", "gray_pt", "maslov", "balauca", "yong"])
    ==================
    Method: gray
    ------------------
    Depth:          50
    CNOT count:     30
    Qubit count:    5
    ==================
    Method: gray_pt
    ------------------
    Depth:          34
    CNOT count:     16
    Qubit count:    5
    ==================
    Method: maslov
    ------------------
    Depth:          43
    CNOT count:     18
    Qubit count:    6
    ==================
    Method: balauca
    ------------------
    Depth:          22
    CNOT count:     18
    Qubit count:    7
    ==================
    Method: yong
    ------------------
    Depth:          77
    CNOT count:     30
    Qubit count:    6
    >>> benchmark_mcx(10, methods = ["gray", "gray_pt", "balauca", "yong"])
    ==================
    Method: gray
    ------------------
    Depth:          3106
    CNOT count:     2046
    Qubit count:    11
    ==================
    Method: gray_pt
    ------------------
    Depth:          2050
    CNOT count:     1024
    Qubit count:    11
    ==================
    Method: balauca
    ------------------
    Depth:          53
    CNOT count:     54
    Qubit count:    18
    ==================
    Method: yong
    ------------------
    Depth:          621
    CNOT count:     264
    Qubit count:    12

    **Mid circuit measurement based methods**

    The ``gidney`` and ``jones`` method are unique in the way that they require
    mid circuit measurements. The measurements are inserted retroactively by the
    :meth:`.compile <qrisp.QuantumSession.compile>` method, because immediate compilation
    would prevent evaluation of the statevector (since a measurement is involved).

    Instead a tentative (measurement free) representative is inserted and replaced
    at compile time.

    To get a better understanding consider the following script:

    >>> from qrisp import QuantumVariable, mcx
    >>> control = QuantumVariable(2)
    >>> target = QuantumBool()
    >>> mcx(control, target, method = "jones")
    >>> print(control.qs)
    QuantumCircuit:
    ---------------
                     ┌───────────────────────────┐
          control.0: ┤0                          ├
                     │                           │
          control.1: ┤1                          ├
                     │                           │
           target.0: ┤2 uncompiled_jones_toffoli ├
                     │                           │
    jones_ancilla.0: ┤3                          ├
                     │                           │
    jones_ancilla.1: ┤4                          ├
                     └───────────────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    QuantumBool target

    We see that there is no classical bit and therefore also no measurement.
    The statevector can still be accessed:

    >>> print(control.qs.statevector())
    |00>*|0>

    To introduce the measurement we simply call the :meth:`.compile <qrisp.QuantumSession.compile>` method
    with the keyword ``compile_mcm = True``:

    >>> qc = control.qs.compile(compile_mcm = True)
    >>> print(qc)
                 ┌─────────────────────────┐
      control.0: ┤0                        ├
                 │                         │
      control.1: ┤1                        ├
                 │                         │
       target.0: ┤2                        ├
                 │  compiled_jones_toffoli │
    workspace_0: ┤3                        ├
                 │                         │
    workspace_1: ┤4                        ├
                 │                         │
           cb_0: ╡0                        ╞
                 └─────────────────────────┘

    Because there is a measurement now, the statevector can no longer be accessed.

    >>> qc.statevector_array()
    Exception: Unitary of operation measure not defined.

    However the T-depth went down by 50%:

    >>> print(qc.t_depth())
    1
    >>> print(control.qs.compile(compile_mcm = False).t_depth())
    2

    A similar construction holds for the `Gidney's temporary logical AND <https://arxiv.org/abs/1709.06648>`_.
    However there are additional details: This technique always comes in pairs. A computation
    and an uncomputation. The computation circuit has a T-depth of 1 and the uncomputation
    circuit has a T-depth of 0. The uncomputation circuit however contains a measurement.

    As you can imagine, this measurement is also inserted at compile time.

    Even though both circuits are not the inverses of each other, Qrisp will use
    the respective partner if called to invert:

    >>> control = QuantumVariable(2)
    >>> target = QuantumBool()
    >>> mcx(control, target, method = "gidney")
    >>> print(control.qs)
    QuantumCircuit:
    ---------------
               ┌────────────────────────┐
    control.0: ┤0                       ├
               │                        │
    control.1: ┤1 uncompiled_gidney_mcx ├
               │                        │
     target.0: ┤2                       ├
               └────────────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    QuantumBool target

    This even works in conjunction with the :ref:`uncomputation module <Uncomputation>`:

    >>> target.uncompute()
    >>> print(target.qs)
    QuantumCircuit:
    ---------------
               ┌────────────────────────┐┌────────────────────────────┐
    control.0: ┤0                       ├┤0                           ├
               │                        ││                            │
    control.1: ┤1 uncompiled_gidney_mcx ├┤1 uncompiled_gidney_mcx_inv ├
               │                        ││                            │
     target.0: ┤2                       ├┤2                           ├
               └────────────────────────┘└────────────────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control

    To introduce the measurement, we call the compile method.

    >>> print(target.qs.compile(compile_mcm = True))
                 ┌──────────────────────┐┌──────────────────────────┐
      control.0: ┤0                     ├┤0                         ├
                 │                      ││                          │
      control.1: ┤1 compiled_gidney_mcx ├┤1                         ├
                 │                      ││  compiled_gidney_mcx_inv │
    workspace_0: ┤2                     ├┤2                         ├
                 └──────────────────────┘│                          │
           cb_0: ════════════════════════╡0                         ╞
                                         └──────────────────────────┘

    Apart from uncomputation, the inverted Gidney mcx can also be accessed via,
    the :ref:`InversionEnvironment`:

    ::

        from qrisp import invert

        control = QuantumVariable(2)
        target = QuantumBool()

        with invert():
            mcx(control, target, method = "gidney")


    >>> print(control.qs)
    QuantumCircuit:
    ---------------
               ┌────────────────────────────┐
    control.0: ┤0                           ├
               │                            │
    control.1: ┤1 uncompiled_gidney_mcx_inv ├
               │                            │
     target.0: ┤2                           ├
               └────────────────────────────┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable control
    QuantumBool target


    """

    from qrisp.misc import bin_rep
    from qrisp.alg_primitives.mcx_algs import (
        GidneyLogicalAND,
        amy_toffoli,
        jones_toffoli,
        jasp_gidney_mcx,
        jasp_gidney_mcx_inv,
        khattar_mcx,
    )
    from qrisp.core import QuantumVariable
    from qrisp.qtypes import QuantumBool

    if isinstance(controls, list):

        new_controls = []
        for qbl in controls:
            if isinstance(qbl, QuantumBool):
                new_controls.append(qbl[0])
            else:
                new_controls.append(qbl)
        controls = new_controls

    if isinstance(target, (list, QuantumVariable)):

        if isinstance(target, QuantumBool):
            target = target[0]
        else:
            raise Exception("mcx target is not of type Qubit or QuantumBool")

    if not check_for_tracing_mode():

        qubits_0 = list(controls)
        qubits_1 = [target]

        n = len(qubits_0)

        if n == 0:
            return controls, target
        elif n == 1:
            method = "gray"
            append_operation(
                std_ops.MCXGate(len(qubits_0), ctrl_state, method=method),
                qubits_0 + qubits_1,
            )
            return

        if not isinstance(ctrl_state, str):
            if ctrl_state == -1:
                ctrl_state += 2**n
            ctrl_state = bin_rep(ctrl_state, n)[::-1]

        if len(ctrl_state) != n:
            raise Exception(
                f"Given control state {ctrl_state} does not match control qubit amount {n}"
            )
    else:
        qubits_0 = controls
        if isinstance(
            qubits_0, (QuantumVariable, DynamicQubitArray)
        ) and method not in ["balauca", "khattar"]:
            method = "balauca"
        qubits_1 = [target]
    from qrisp.alg_primitives.mcx_algs import (
        balauca_dirty,
        balauca_mcx,
        hybrid_mcx,
        maslov_mcx,
        yong_mcx,
        jasp_balauca_mcx,  # Merge the import
    )

    if method in ["gray", "gray_pt", "gray_pt_inv"]:
        append_operation(
            std_ops.MCXGate(len(qubits_0), ctrl_state, method=method),
            qubits_0 + qubits_1,
        )
    elif method == "gidney":
        if len(qubits_0) != 2:
            raise Exception(
                f"Tried to call Gidney logical AND with {len(qubits_0)} controls instead of two"
            )

        if check_for_tracing_mode():
            jasp_gidney_mcx(qubits_0[0], qubits_0[1], qubits_1[0])
        else:
            append_operation(
                GidneyLogicalAND(ctrl_state=ctrl_state),
                qubits_0 + qubits_1,
            )

    elif method == "gidney_inv":
        if len(qubits_0) != 2:
            raise Exception(
                f"Tried to call Gidney logical AND with {len(qubits_0)} controls instead of two"
            )

        if check_for_tracing_mode():
            jasp_gidney_mcx_inv(qubits_0[0], qubits_0[1], qubits_1[0])
        else:
            append_operation(
                GidneyLogicalAND(ctrl_state=ctrl_state, inv=True),
                qubits_0 + qubits_1,
            )

    elif method == "maslov":
        from qrisp import QuantumBool

        if n >= 3:
            ancilla = [QuantumBool(name="maslov_anc_")]
        else:
            ancilla = []
        append_operation(maslov_mcx(n, ctrl_state), qubits_0 + ancilla + qubits_1)

        [qv.delete() for qv in ancilla]

    elif method == "balauca":
        if check_for_tracing_mode():
            jasp_balauca_mcx(qubits_0, qubits_1, ctrl_state)
        else:
            balauca_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state)

    elif method == "balauca_dirty":
        balauca_dirty(qubits_0, qubits_1, k=num_ancilla, ctrl_state=ctrl_state)

    elif method == "yong":
        yong_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state)

    elif method == "hybrid":
        hybrid_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state, num_ancilla=num_ancilla)

    elif method == "amy":
        if len(qubits_0) != 2:
            raise Exception(
                f"Tried to call Amy MCX with {len(qubits_0)} controls instead of two"
            )
        amy_toffoli(qubits_0, qubits_1, ctrl_state=ctrl_state)

    elif method == "jones":
        if len(qubits_0) != 2:
            raise Exception(
                f"Tried to call Jones MCX with {len(qubits_0)} controls instead of two"
            )
        jones_toffoli(qubits_0, qubits_1, ctrl_state=ctrl_state)

    elif method == "auto":
        # if n <= 3:
        #     return mcx(qubits_0, qubits_1, method = "gray", ctrl_state = ctrl_state)
        # if 3 < n < 5:
        #     return mcx(qubits_0, qubits_1, method = "maslov", ctrl_state = ctrl_state)
        # else:
        #     return mcx(qubits_0, qubits_1, method = "balauca", ctrl_state = ctrl_state) # noqa:501
        gate = std_ops.MCXGate(len(qubits_0), ctrl_state, method="auto")
        append_operation(gate, qubits_0 + qubits_1)

    elif method == "khattar":
        # if check_for_tracing_mode():
        khattar_mcx(qubits_0, qubits_1, ctrl_state)
        # else:
        #    balauca_mcx(qubits_0, qubits_1, ctrl_state=ctrl_state) # PLACEHOLDER

    else:
        raise Exception(f"Don't know mcx method {method}")

    return controls, target


def mcz(qubits, method="auto", ctrl_state=-1, num_ancilla=1):
    """
    Applies a multi-controlled Z gate.

    For more information on the available methods, check
    :meth:`the mcx documentation page <qrisp.mcx>`.

    Parameters
    ----------
    qubits : QuantumVariable or list[Qubits]
        The Qubits to control on.
    method : str, optional
        The synthesis method. Available are ``auto``, ``gray``, ``gray_pt``,
        ``gray_pt_inv``, ``maslov``, ``balauca``, ``yong`` and ``hybrid``. The default
        is ``auto``.
    ctrl_state : int or str, optional
        The state on which to activate the Z gate. The default is "1111...".
    num_ancilla : int, optional
        Specifies the amount of ancilla qubits to use. This parameter is used only if
        the method is set to ``hybrid``. The default is 1.

    """

    from qrisp.misc import gate_wrap

    @gate_wrap(permeability="full", is_qfree=True, name="anc supported mcz")
    def mcz_inner(qubits, method="auto", ctrl_state=-1):
        if len(ctrl_state) != n:
            raise Exception(
                f"Given control state {ctrl_state} does not match"
                f"control qubit amount {n}"
            )

        from qrisp import h, x

        if ctrl_state[-1] == "0":
            x(qubits[-1])

        h(qubits[-1])
        mcx(qubits[:-1], qubits[-1], method=method, ctrl_state=ctrl_state[:-1])
        h(qubits[-1])

        if ctrl_state[-1] == "0":
            x(qubits[-1])

        return qubits

    n = len(qubits)

    from qrisp import bin_rep

    if not isinstance(ctrl_state, str):
        if ctrl_state == -1:
            ctrl_state += 2**n
        ctrl_state = bin_rep(ctrl_state, n)

    if method in ["gray", "auto"]:
        if ctrl_state[-1] == "0":
            x(qubits[-1])

        if len(qubits) > 1:
            append_operation(
                std_ops.ZGate().control(
                    len(qubits) - 1, method=method, ctrl_state=ctrl_state[:-1]
                ),
                qubits,
            )
        else:
            z(qubits[0])

        if ctrl_state[-1] == "0":
            x(qubits[-1])

        return qubits

    return mcz_inner(qubits, method, ctrl_state)


def mcp(phi, qubits, method="auto", ctrl_state=-1):
    """
    Applies a multi-controlled phase gate.

    The available methods are:

    * ``gray`` , which performs a traversal of the gray code.

    * ``balauca`` , which is a modified version of the algorithm presented `here
      <https://www.iccs-meeting.org/archive/iccs2022/papers/133530169.pdf>`_.

    * ``auto`` , which picks ``gray`` for any qubit count less than 4 and ``balauca``
      otherwise.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits : list[Qubit] or QuantumVariable
        The qubits to apply the multi-controlled phase gate on.
    method : str, optional
        The method to deploy. The default is "auto".
    ctrl_state : str or int, optional
        The control state on which to apply the phase. The default is "111...".

    """

    from qrisp.alg_primitives.mcx_algs import hybrid_mcx, jasp_balauca_mcp, khattar_mcp
    from qrisp import QuantumBool
    from qrisp.misc import bin_rep, gate_wrap
    from qrisp.environments import control
    import numpy as np

    @gate_wrap(permeability="full", is_qfree=True, name="anc supported mcp")
    def balauca_mcp(phi, qubits, ctrl_state):
        from qrisp.circuit.quantum_circuit import convert_to_qb_list

        qubits = convert_to_qb_list(qubits)

        temp = QuantumBool()
        hybrid_mcx(
            qubits,
            temp[0],
            ctrl_state=ctrl_state,
            phase=phi,
            num_ancilla=np.inf,
            use_mcm=True,
        )

        temp.delete()

    n = jlen(qubits)
    if not check_for_tracing_mode():
        if not isinstance(ctrl_state, str):
            if ctrl_state == -1:
                ctrl_state += 2**n
            ctrl_state = bin_rep(ctrl_state, n)[::-1]
    else:
        if method == "auto":
            method = "balauca"
        if isinstance(ctrl_state, str):
            ctrl_state = int(ctrl_state, 2)

    if method == "gray" or method == "gray_pt":
        if ctrl_state[-1] == "0":
            x(qubits[-1])

        if check_for_tracing_mode():
            mcp_gate = std_ops.PGate(sympy.Symbol("alpha")).control(
                n - 1, ctrl_state=ctrl_state[:-1], method=method
            )
            append_operation(mcp_gate, qubits, param_tracers=[phi])
        else:
            mcp_gate = std_ops.PGate(phi).control(
                n - 1, ctrl_state=ctrl_state[:-1], method=method
            )
            append_operation(mcp_gate, qubits)

        if ctrl_state[-1] == "0":
            x(qubits[-1])
        return qubits

    elif method == "balauca":
        if check_for_tracing_mode():
            jasp_balauca_mcp(phi, qubits, ctrl_state)
            return
        else:
            balauca_mcp(phi, qubits, ctrl_state=ctrl_state)
        return qubits

    elif method == "khattar":
        khattar_mcp(phi, qubits, ctrl_state=ctrl_state)
        # return qubits   #is it needed? maybe for static case

    elif method == "auto":
        with control(n < 4):
            mcp(phi, qubits, method="gray", ctrl_state=ctrl_state)
        with control(n >= 4):
            mcp(phi, qubits, method="balauca", ctrl_state=ctrl_state)
        return
    else:
        raise Exception(f"Don't know method {method}")


def p(phi, qubits):
    """
    Applies a phase gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit on which to apply the phase gate.

    """

    if check_for_tracing_mode():
        append_operation(
            std_ops.PGate(sympy.Symbol("alpha")), [qubits], param_tracers=[phi]
        )
    else:
        append_operation(std_ops.PGate(phi), [qubits])

    # std_ops.PGate(phi).append([qubits])

    return qubits


def cp(phi, qubits_0, qubits_1):
    """
    Applies a controlled phase gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit.

    """

    if check_for_tracing_mode():
        cp_gate = std_ops.CPGate(sympy.Symbol("alpha"))
        append_operation(cp_gate, [qubits_0, qubits_1], param_tracers=[phi])
    else:
        cp_gate = std_ops.CPGate(phi)
        append_operation(cp_gate, [qubits_0, qubits_1])

    return qubits_0, qubits_1


def rx(phi, qubits):
    """
    Applies an RX gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the RX gate on.

    """

    if check_for_tracing_mode():
        append_operation(
            std_ops.RXGate(sympy.Symbol("alpha")), [qubits], param_tracers=[phi]
        )
    else:
        append_operation(std_ops.RXGate(phi), [qubits])

    return qubits


def ry(phi, qubits):
    """
    Applies an RY gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the RY gate on.

    """

    if check_for_tracing_mode():
        append_operation(
            std_ops.RYGate(sympy.Symbol("alpha")), [qubits], param_tracers=[phi]
        )
    else:
        append_operation(std_ops.RYGate(phi), [qubits])

    return qubits


def rz(phi, qubits):
    """
    Applies an RZ gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the RZ gate on.

    """

    if check_for_tracing_mode():
        append_operation(
            std_ops.RZGate(sympy.Symbol("alpha")), [qubits], param_tracers=[phi]
        )
    else:
        append_operation(std_ops.RZGate(phi), [qubits])

    return qubits


def crz(phi, qubits_0, qubits_1):
    """
    Applies controlled RZ gate

    Parameters
    ----------
    phi : float or sympy.Symbol
        The angle parameter.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit.

    """

    if check_for_tracing_mode():
        crz_gate = std_ops.RZGate(sympy.Symbol("alpha")).control(1)
        append_operation(crz_gate, [qubits_0, qubits_1], param_tracers=[phi])
    else:
        crz_gate = std_ops.RZGate(phi).control(1)
        append_operation(crz_gate, [qubits_0, qubits_1])

    return qubits_0, qubits_1


def s(qubits):
    """
    Applies an S gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the S gate on.
    """

    append_operation(std_ops.SGate(), [qubits])
    return qubits


def t(qubits):
    """
    Applies a T gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the T gate on.
    """
    append_operation(std_ops.TGate(), [qubits])
    return qubits


def s_dg(qubits):
    """
    Applies a daggered S gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the daggered S gate on.
    """

    append_operation(std_ops.SGate().inverse(), [qubits])
    return qubits


def t_dg(qubits):
    """
    Applies a daggered T gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the daggered T gate on.
    """
    append_operation(std_ops.TGate().inverse(), [qubits])
    return qubits


def sx(qubits):
    """
    Applies an SX gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the SX gate on.
    """

    append_operation(std_ops.SXGate(), [qubits])

    return qubits


def sx_dg(qubits):
    """
    Applies a daggered SX gate.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the daggered SX gate on.
    """
    append_operation(std_ops.SXDGGate().inverse(), [qubits])
    return qubits


def gphase(phi, qubits):
    """
    Applies a global phase. This gate turns into a phase gate when controlled.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The global phase to apply.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the global phase gate on.
    """

    if check_for_tracing_mode():
        append_operation(
            std_ops.GPhaseGate(sympy.Symbol("alpha")), [qubits], param_tracers=[phi]
        )
    else:
        append_operation(std_ops.GPhaseGate(phi), [qubits])
    return qubits


def xxyy(phi, beta, qubits_0, qubits_1):
    """
    Applies an XXYY interaction gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The global phase to apply.
    beta : float or sympy.Symbol
        The other angle parameter.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit to perform the XXYY gate on.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit to perform the XXYY gate on.
    """

    if check_for_tracing_mode():
        xxyy_gate = std_ops.XXYYGate(sympy.Symbol("alpha"), sympy.Symbol("beta"))
        append_operation(xxyy_gate, [qubits_0, qubits_1], param_tracers=[phi, beta])

    else:
        xxyy_gate = std_ops.XXYYGate(phi, beta)
        append_operation(xxyy_gate, [qubits_0, qubits_1])

    return qubits_0, qubits_1


def rzz(phi, qubits_0, qubits_1):
    """
    Applies an RZZ gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first argument to perform the RZZ gate one.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second argument to perform the RZZ gate one.
    """

    if check_for_tracing_mode():
        rzz_gate = std_ops.RZZGate(sympy.Symbol("alpha"))
        append_operation(rzz_gate, [qubits_0, qubits_1], param_tracers=[phi])
    else:
        rzz_gate = std_ops.RZZGate(phi)
        append_operation(rzz_gate, [qubits_0, qubits_1])
    return qubits_0, qubits_1


def rxx(phi, qubits_0, qubits_1):
    """
    Applies an RXX gate.

    Parameters
    ----------
    phi : float or sympy.Symbol
        The phase to apply.
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first argument to perform the RXX gate one.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second argument to perform the RXX gate one.
    """

    if check_for_tracing_mode():
        rxx_gate = std_ops.RXXGate(sympy.Symbol("alpha"))
        append_operation(rxx_gate, [qubits_0, qubits_1], param_tracers=[phi])
    else:
        rxx_gate = std_ops.RXXGate(phi)
        append_operation(rxx_gate, [qubits_0, qubits_1])
    return qubits_0, qubits_1


def u3(theta, phi, lam, qubits):
    """
    Applies an U3 gate.

    Parameters
    ----------
    theta : float or sympy.Symbol
        The first angle parameter.
    phi : float or sympy.Symbol
        The second angle parameter.
    lambda : float or sympy.Symbol
        The third angle parameter.
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to perform the U3 gate on.
    """

    if check_for_tracing_mode():
        append_operation(
            std_ops.U3Gate(
                sympy.Symbol("alpha"), sympy.Symbol("beta"), sympy.Symbol("gamma")
            ),
            [qubits],
            param_tracers=[theta, phi, lam],
        )
    else:
        append_operation(std_ops.U3Gate(theta, phi, lam), [qubits])

    return qubits


def measure(qubits):
    """
    Performs a measurement of the specified Qubit.

    Parameters
    ----------
    qubit : Qubit or list[Qubit] or QuantumVariable
        The Qubit to measure.
    clbit : Clbit, optional
        The Clbit to store the result in. By default, a new Clbit will be created.

    """
    from qrisp import find_qs
    from qrisp.jasp import TracingQuantumSession

    qs = find_qs(qubits)

    if not isinstance(qs, TracingQuantumSession):

        clbits = []
        if hasattr(qubits, "__len__"):
            for qb in qubits:
                try:
                    clbits.append(qs.add_clbit())
                except AttributeError:
                    clbits.append(qs.add_clbit())

        else:
            clbits = qs.add_clbit()

        append_operation(std_ops.Measurement(), [qubits], [clbits])

        return clbits
    else:
        from qrisp.jasp import (
            Measurement_p,
            AbstractQubit,
            AbstractQubitArray,
            DynamicQubitArray,
        )
        from qrisp import QuantumVariable, QuantumArray
        
        if not qs.abs_qc._trace is jax.core.trace_ctx.trace:
            raise Exception(
                """Lost track of QuantumCircuit during tracing. This might have been caused by a missing quantum_kernel decorator or not using quantum prefix control (like q_fori_loop, q_cond). Please visit https://www.qrisp.eu/reference/Jasp/Quantum%20Kernel.html for more details"""
            )

        if isinstance(qubits, (DynamicQubitArray, QuantumVariable, QuantumArray)):
            res = qubits.measure()
        elif isinstance(qubits, jax.core.Tracer) and isinstance(qubits.aval, (AbstractQubitArray, AbstractQubit)):
            res, abs_qc = Measurement_p.bind(qubits, qs.abs_qc)
            qs.abs_qc = abs_qc
        else:
            raise Exception(f"Tried to measure type {type(qubits)}")

        return res


def reset(qubits):
    """
    Performs a reset of the specified Qubit.

    Parameters
    ----------
    qubit : Qubit or list[Qubit] or QuantumVariable
        The Qubit to measure.
    """
    from qrisp import find_qs
    from qrisp.jasp import TracingQuantumSession

    qs = find_qs(qubits)

    if not isinstance(qs, TracingQuantumSession):
        append_operation(std_ops.Reset(), [qubits])
        return None
    else:
        from qrisp.jasp import reset_p, AbstractQubit, AbstractQubitArray
        from qrisp import QuantumVariable

        if isinstance(qubits, QuantumVariable):
            abs_qc = reset_p.bind(qubits.reg.tracer, qs.abs_qc)
        elif isinstance(qubits.aval, AbstractQubitArray):
            abs_qc = reset_p.bind(qubits.tracer, qs.abs_qc)
        elif isinstance(qubits.aval, AbstractQubit):
            abs_qc = reset_p.bind(qubits, qs.abs_qc)
        else:
            raise Exception(f"Tried to reset type {type(qubits.aval)}")

        qs.abs_qc = abs_qc

        return None


def barrier(qubits):
    """
    A visual marker for structuring the QuantumCircuit.

    Parameters
    ----------
    qubits : Qubit or list[Qubit] or QuantumVariable
        The Qubit to apply the barrier to.

    Examples
    --------

    >>> from qrisp import QuantumVariable, x, y, barrier
    >>> qv = QuantumVariable(5)
    >>> x(qv)
    >>> barrier(qv)
    >>> y(qv)
    >>> print(qv.qs)

    .. code-block:: none

        QuantumCircuit:
        --------------
              ┌───┐ ░ ┌───┐
        qv.0: ┤ X ├─░─┤ Y ├
              ├───┤ ░ ├───┤
        qv.1: ┤ X ├─░─┤ Y ├
              ├───┤ ░ ├───┤
        qv.2: ┤ X ├─░─┤ Y ├
              ├───┤ ░ ├───┤
        qv.3: ┤ X ├─░─┤ Y ├
              ├───┤ ░ ├───┤
        qv.4: ┤ X ├─░─┤ Y ├
              └───┘ ░ └───┘
        Live QuantumVariables:
        ---------------------
        QuantumVariable qv

    """
    from qrisp import Qubit

    if isinstance(qubits, Qubit):
        qubits = [qubits]

    append_operation(std_ops.Barrier(len(qubits)), qubits)

    return qubits


def swap(qubits_0, qubits_1):
    """
    Applies a SWAP gate.

    Parameters
    ----------
    qubits_0 : Qubit or list[Qubit] or QuantumVariable
        The first Qubit.
    qubits_1 : Qubit or list[Qubit] or QuantumVariable
        The second Qubit.

    """
    append_operation(std_ops.SwapGate(), [qubits_0, qubits_1])

    return qubits_0, qubits_1



"""
********************************************************************************
* Copyright (c) 2024 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import math
import threading

import numpy as np
from numba import njit

from qrisp.core import QuantumVariable, QuantumArray
from qrisp.core.compilation import qompiler
from qrisp.interface import BatchedBackend


def get_measurement(
    hamiltonian,
    qarg,
    precision=0.01,
    backend=None,
    compile=True,
    compilation_kwargs={},
    subs_dic={},
    precompiled_qc=None,
    diagonalisation_method="commuting_qw",
    measurement_data=None,  # measurement settings
):
    r"""
    This method returns the expected value of a Hamiltonian for the state of a quantum argument.

    Parameters
    ----------
    qarg : :ref:`QuantumVariable` or list[Qubit]
        The quantum argument to evaluate the Hamiltonian on.
    precision: float, optional
        The precision with which the expectation of the Hamiltonian is to be evaluated.
        The default is 0.01. The number of shots scales quadratically with the inverse precision.
    backend : :ref:`BackendClient`, optional
        The backend on which to evaluate the quantum circuit. The default can be
        specified in the file default_backend.py.
    compile : bool, optional
        Boolean indicating if the .compile method of the underlying QuantumSession
        should be called before. The default is ``True``.
    compilation_kwargs  : dict, optional
        Keyword arguments for the compile method. For more details check
        :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default
        is ``{}``.
    subs_dic : dict, optional
        A dictionary of Sympy symbols and floats to specify parameters in the case
        of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`.
        The default is ``{}``.
    precompiled_qc : QuantumCircuit, optional
            A precompiled quantum circuit.
    diagonalisation_method : str, optional
        Specifies the method for grouping and diagonalizing the QubitOperator.
        Available are ``commuting_qw``, i.e., the operator is grouped based on qubit-wise commutativity of terms,
        and ``commuting``, i.e., the operator is grouped based on commutativity of terms.
        The default is ``commuting_qw``.
    measurement_data : QubitOperatorMeasurement
        Cached data to accelerate the measurement procedure. Automatically generated by default.

    Raises
    ------
    Exception
        If the containing QuantumSession is in a quantum environment, it is not
        possible to execute measurements.

    Returns
    -------
    float
        The expected value of the Hamiltonian.

    """

    from qrisp import QuantumSession, merge

    if isinstance(qarg, QuantumVariable):
        if qarg.is_deleted():
            raise Exception("Tried to get measurement from deleted QuantumVariable")
        qs = qarg.qs

    elif isinstance(qarg, QuantumArray):
        for qv in qarg.flatten():
            if qv.is_deleted():
                raise Exception(
                    "Tried to measure QuantumArray containing deleted QuantumVariables"
                )
        qs = qarg.qs
    elif isinstance(qarg, list):
        qs = QuantumSession()
        for arg in qarg:
            if isinstance(arg, QuantumVariable) and qv.is_deleted():
                raise Exception(
                    "Tried to measure QuantumArray containing deleted QuantumVariables"
                )
            merge(qs, arg)

    if backend is None:
        if qs.backend is None:
            from qrisp.default_backend import def_backend

            backend = def_backend
        else:
            backend = qarg.qs.backend

    if len(qs.env_stack) != 0:
        raise Exception("Tried to get measurement within open environment")

    hamiltonian = hamiltonian.hermitize()
    hamiltonian = hamiltonian.eliminate_ladder_conjugates()
    hamiltonian = hamiltonian.apply_threshold(0)
    if len(hamiltonian.terms_dict) == 0:
        return 0

    # Copy circuit in over to prevent modification
    if precompiled_qc is None:
        if compile:
            qc = qompiler(qs, **compilation_kwargs)
        else:
            qc = qs.copy()
        qubit_list = qarg
    else:
        qc = precompiled_qc.copy()
        qubit_list = qc.qubits[: len(qarg.reg)]

    # Bind parameters
    if subs_dic:
        qc = qc.bind_parameters(subs_dic)
        from qrisp.core.compilation import combine_single_qubit_gates

        qc = combine_single_qubit_gates(qc)

    qc = qc.transpile()

    if measurement_data is None:
        measurement_data = QubitOperatorMeasurement(
            hamiltonian, diagonalisation_method=diagonalisation_method
        )

    return measurement_data.get_measurement(qc, qubit_list, precision, backend)


class QubitOperatorMeasurement:

    def __init__(self, hamiltonian, diagonalisation_method="commuting_qw"):

        n = hamiltonian.find_minimal_qubit_amount()

        if diagonalisation_method == "commuting_qw":
            temp_groups = hamiltonian.commuting_qw_groups()
            self.groups = []
            # In order for the change of basis function (below) to work properly,
            # the ladder terms either need to completely agree or completely disagree
            for group in temp_groups:
                self.groups.extend(
                    group.group_up(
                        lambda a, b: a.ladders_agree(b) or not a.ladders_intersect(b)
                    )
                )

        elif diagonalisation_method == "commuting":
            temp_groups = hamiltonian.group_up(lambda a, b: a.commute_pauli(b))
            self.groups = []
            # In order for the change of basis function (below) to work properly,
            # the ladder terms either need to completely agree or completely disagree
            for group in temp_groups:
                self.groups.extend(
                    group.group_up(
                        lambda a, b: a.ladders_agree(b) or not a.ladders_intersect(b)
                    )
                )

        else:
            raise Exception(
                f"Unknown diagonalisation method: {diagonalisation_method}."
            )

        self.stds = []
        self.change_of_basis_gates = []
        self.measurement_operators = []

        for group in self.groups:

            qv = QuantumVariable(n)

            meas_op = group.change_of_basis(qv, diagonalisation_method)
            self.change_of_basis_gates.append(qv.qs.to_gate())
            self.measurement_operators.append(meas_op)

            # Collect standard deviation
            self.stds.append(np.sqrt(meas_op.get_operator_variance(n=n)))

        N = sum(self.stds)
        self.shots_list = [N * s for s in self.stds]

    def get_measurement(self, qc, qubit_list, precision, backend):

        from qrisp.misc import get_measurement_from_qc

        results = [0]*len(self.measurement_operators)
        meas_coeffs = list(results)
        meas_ops = list(results)
        
        def measurement_thread(i):
            
            group = self.measurement_operators[i]

            shots = int(self.shots_list[i] / precision**2)

            qubits = [
                qubit_list[j] for j in range(self.change_of_basis_gates[i].num_qubits)
            ]

            curr = qc.copy()
            curr.append(self.change_of_basis_gates[i], qubits)

            res = get_measurement_from_qc(curr, list(qubit_list), backend, shots)
            results[i] = res

            temp_meas_ops = []
            temp_coeff = []
            for term, coeff in group.terms_dict.items():
                temp_meas_ops.append(term.serialize())
                temp_coeff.append(coeff)

            meas_coeffs[i] = temp_coeff
            meas_ops[i] = temp_meas_ops
        
        
        threads = []
        
        for i in range(len(self.measurement_operators)):
            
            if isinstance(backend, BatchedBackend):
                thread = threading.Thread(target = measurement_thread, args = (i,))
                thread.start()
                threads.append(thread)
            else:
                measurement_thread(i)
        
        if isinstance(backend, BatchedBackend):
            backend.dispatch(min_calls = len(self.measurement_operators))
            for thread in threads:
                thread.join()

        samples = create_padded_array([list(res.keys()) for res in results]).astype(
            np.int64
        )
        probs = create_padded_array([list(res.values()) for res in results])
        meas_ops = create_padded_array(meas_ops, use_tuples=True).astype(np.int64)
        meas_coeffs = create_padded_array(meas_coeffs)

        return evaluate_expectation_jitted(samples, probs, meas_ops, meas_coeffs)


def create_padded_array(list_of_lists, use_tuples=False):
    """
    Create a padded numpy array from a list of lists with varying lengths.

    Parameters:
    list_of_lists (list): A list of lists with potentially different lengths.

    Returns:
    numpy.ndarray: A 2D numpy array with padded rows.
    """
    # Find the maximum length of any list in the input
    max_length = max(len(lst) for lst in list_of_lists)

    # Create a padded list of lists
    if not use_tuples:
        padded_lists = [lst + [0] * (max_length - len(lst)) for lst in list_of_lists]
    else:
        padded_lists = [
            lst + [(0, 0, 0, 0)] * (max_length - len(lst)) for lst in list_of_lists
        ]

    # Convert to numpy array
    return np.array(padded_lists)


#
# Evaluate expectation
#


def evaluate_expectation(samples, probs, operators, coefficients):
    """
    Evaluate the expectation.

    """
    # print(results)
    # print(operators)
    # print(coefficients)
    # raise

    expectation = 0

    for index1, ops in enumerate(operators):
        for index2, op in enumerate(ops):
            for i in range(len(samples[index1])):
                outcome, probability = samples[index1, i], probs[index1, i]
                expectation += (
                    probability
                    * evaluate_observable(op, outcome)
                    * np.real(coefficients[index1][index2])
                )

    return expectation


def evaluate_observable(observable: tuple, x: int):
    # This function evaluates how to compute the energy of a measurement sample x.
    # Since we are also considering ladder operators, this energy can either be
    # 0, -1 or 1. For more details check out the comments of QubitOperator.get_conjugation_circuit

    # The observable is given as tuple, containing for integers and a boolean.
    # To understand the meaning of these integers check QubitTerm.serialize.

    # Unwrap the tuple
    z_int, AND_bits, AND_ctrl_state, contains_ladder = observable

    # Compute whether the sign should be sign flipped based on the Z operators
    sign_flip_int = z_int & x
    sign_flip = 0
    while sign_flip_int:
        sign_flip += sign_flip_int & 1
        sign_flip_int >>= 1

    # If there is a ladder operator in the term, we need to half the energy
    # because we want to measure (|110><110| - |111><111|)/2
    if contains_ladder:
        prefactor = 0.5
    else:
        prefactor = 1

    # If there are no and bits, we return the result
    if AND_bits == 0:
        return prefactor * (-1) ** sign_flip

    # Otherwise we apply the AND_ctrl_state to flip the appropriate bits.
    corrected_x = x ^ AND_ctrl_state

    # If all bits are in the 0 state the AND is true.
    if corrected_x & AND_bits == 0:
        return prefactor * (-1) ** sign_flip
    else:
        return 0


evaluate_observable_jitted = njit(cache=True)(evaluate_observable)


@njit(cache=True)
def evaluate_expectation_jitted(samples, probs, operators, coefficients):
    """
    Evaluate the expectation.

    """
    expectation = 0

    for index1, ops in enumerate(operators):
        for index2, op in enumerate(ops):
            for i in range(len(samples[index1])):
                outcome, probability = samples[index1, i], probs[index1, i]
                expectation += (
                    probability
                    * evaluate_observable_jitted(op, outcome)
                    * np.real(coefficients[index1][index2])
                )

    return expectation


def partition(values, num_qubits):
    """
    Partitions a list of integers into a list of lists of integers with size 64 bit.

    Parameters
    ----------
    values : list[int]
        A list of integers.
    num_qubits : int
        The maximal number of bits.

    Returns
    -------
    partition : list[np.array]
        A list of NumPy numpy.uint64 arrays.

    """

    M = math.ceil(num_qubits / 64)
    N = len(values)

    zeros = [[0] * N for k in range(M)]
    partition = [values]
    partition.extend(zeros)

    lower_mask = (1 << 64) - 1

    for j in range(1, M):
        for i in range(N):
            partition[j][i] = partition[j - 1][i] & lower_mask
            partition[j + 1][i] = partition[j - 1][i] >> 64

    if M == 1:
        return [np.array(partition[0], dtype=np.uint64)]
    else:
        return [np.array(part, dtype=np.uint64) for part in partition[1:]]



"""
********************************************************************************
* Copyright (c) 2025 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import qiskit
from qrisp.interface import VirtualBackend, BatchedBackend


class AQTBackend(BatchedBackend):
    """
    This class instantiates a :ref:`VirtualBackend` using an AQT backend. 
    This allows easy access to AQT backends through the qrisp interface.

    Using this backend requires addtional dependencies: ``pip install qiskit-aqt-provider``. 

    Parameters
    ----------
    api_token : str
        An API token for `AQT ARNICA <https://www.aqt.eu/products/arnica/>`_.
    device_instance : str
        The device instance of the AQT backend such as ``ibex`` or ``simulator_noise``.
        For an up-to-date list, see the AQT ARNICA website.
    workspace : str 
        The workspace for a company or project.

    Examples
    --------

    We evaluate a :ref:`QuantumFloat` multiplication on the 12-qubit AQT IBEX.

    >>> from qrisp import QuantumFloat
    >>> from qrisp.interface import AQTBackend
    >>> # qrisp_ibex = AQTBackend(api_token = "YOUR_AQT_ARNICA_TOKEN", device_instance = "simulator_noise", workspace = "aqt_simulators")
    >>> qrisp_ibex = AQTBackend(api_token = "YOUR_AQT_ARNICA_TOKEN", device_instance = "ibex", workspace = "YOUR_COMPANY_OR_PROJECT_NAME")
    >>> a = QuantumFloat(2)
    >>> a[:] = 2
    >>> b = a*a
    >>> b.get_measurement(backend = qrisp_ibex, shots = 100)
    {4: 0.49,
    8: 0.11,
    2: 0.08,
    0: 0.06,
    14: 0.06,
    5: 0.04,
    12: 0.04,
    13: 0.03,
    3: 0.02,
    10: 0.02,
    15: 0.02,
    6: 0.01,
    7: 0.01,
    11: 0.01}

    """

    def __init__(self, api_token, device_instance, workspace):

        if not isinstance(api_token, str):
            raise TypeError(
                "api_token must be a string. You can create an API token on the AQT ARNICA website."
            )
        
        if not isinstance(workspace, str):
            raise TypeError(
                "workspace must be a string."
            )

        if not isinstance(device_instance, str):
            raise TypeError(
                "Please provide a device_instance as a string. You can retrieve a list of available devices on the AQT ARNICA website."
            )

        try:
            from qiskit_aqt_provider import AQTProvider
            from qiskit_aqt_provider.primitives import AQTSampler
        except ImportError:
            raise ImportError(
                "Please install qiskit-aqt-provider to use the AQTBackend. You can do this by running `pip install qiskit-aqt-provider`."
            )
        
        provider = AQTProvider(api_token)
        backend = provider.get_backend(name = device_instance, workspace = workspace)

        """
        def run(qasm_str, shots=None, token=""):
            if shots is None:
                shots = 100

            # Convert to qiskit
            qiskit_qc = qiskit.QuantumCircuit.from_qasm_str(qasm_str)

            # Make circuit with one monolithic register
            new_qiskit_qc = qiskit.QuantumCircuit(len(qiskit_qc.qubits), len(qiskit_qc.clbits))
            for instr in qiskit_qc:
                new_qiskit_qc.append(
                    instr.operation,
                    [qiskit_qc.qubits.index(qb) for qb in instr.qubits],
                    [qiskit_qc.clbits.index(cb) for cb in instr.clbits],
                )
            
            # Instantiate a sampler on the execution backend.
            sampler = AQTSampler(backend)

            # Optional: set the transpiler's optimization level.
            # Optimization level 3 typically provides the best results.
            sampler.set_transpile_options(optimization_level=3)

            # Sample the circuit on the execution backend.
            result = sampler.run(new_qiskit_qc, shots=shots).result()

            quasi_dist = result.quasi_dists[0]
            
            # Format to fit the qrisp result format
            result_dic = {}

            for item in list(quasi_dist.keys()):
                new_key = bin(item)[2:].zfill(len(qiskit_qc.clbits))
                result_dic.setdefault(new_key, quasi_dist[item])

            return result_dic
        """

        def run_batch_aqt(batch):
        
            circuit_batch = []
            shot_batch = []
            cl_bits_batch = []
            for qc, shots in batch:
                # Sometimes wrong results without transpilation 
                qiskit_qc = qc.transpile().to_qiskit()

                # Make circuit with one monolithic register
                new_qiskit_qc = qiskit.QuantumCircuit(len(qiskit_qc.qubits), len(qiskit_qc.clbits))
                for instr in qiskit_qc:
                    new_qiskit_qc.append(
                        instr.operation,
                        [qiskit_qc.qubits.index(qb) for qb in instr.qubits],
                        [qiskit_qc.clbits.index(cb) for cb in instr.clbits],
                    )

                circuit_batch.append(new_qiskit_qc)
                cl_bits_batch.append(len(qiskit_qc.clbits))

                if shots is None:
                    shots = 100
            
                shot_batch.append(shots)

            # Instantiate a sampler on the execution backend.
            sampler = AQTSampler(backend)

            # Optional: set the transpiler's optimization level.
            # Optimization level 3 typically provides the best results.
            sampler.set_transpile_options(optimization_level=3)

            # Sample the circuit on the execution backend.
            results = sampler.run(circuit_batch, shots = max(shot_batch)).result()

            quasi_dist_batch = []
            for i in range(len(batch)):
                quasi_dist = results.quasi_dists[i]
                cl_bits = cl_bits_batch[i]

                new_quasi_dist = {}
                for item in list(quasi_dist.keys()):
                    new_key = bin(item)[2:].zfill(cl_bits)
                    new_quasi_dist.setdefault(new_key, quasi_dist[item])

                quasi_dist_batch.append(new_quasi_dist)
    
            return quasi_dist_batch

        # Call BatchedBackend constructor
        if isinstance(backend.name, str):
            name = backend.name
        else:
            name = backend.name()

        super().__init__(run_batch_aqt)